//----------------------------------------------------------------------------------
// Module D17 par Jean-Michel Dubois
// Train management
//----------------------------------------------------------------------------------

#include "config.h"

#if PROTOCOLE == D17PPV3

#include "block.h"
#include "dcc.h"
#include "slot.h"

Train::Train()
{
	// constantes
	m_adresse = 0;
	m_pasMax = 0;  // pas DCC de vitesse maxi
	m_pas320 = 0;	// pas DCC 320 km/h
	m_pas300 = 0;
	m_pas220 = 0;
	m_pas160 = 0;  // pas DCC 160 km/h
	m_pas120 = 0;
	m_pas100 = 0;
	m_pas90 = 0;   // pas DCC 90 km/h
	m_pas60 = 0;   // pas DCC 60 km/h
	m_pas30 = 0;   // pas DCC 30 km/h
	m_pas15 = 0;   // pas DCC 30 km/h
	m_pasDCC = 28;
	m_vmanoeuvre = 0;  // numéro de fonction de la vitesse de manoeuvre
	// variables
	m_zone = 255;		// zone actuelle = AUCUNE
	m_sensAvant = 0;	// sens en marche avant : SANS_SENS, SENS_PAIR, SENS_IMPAIR
	m_sensActuel = 0;	// sens de marche actuel
	m_pasSouris = 0;     // pas DCC de consigne de la souris
	m_pasActuel = 0; // pas DCC actuel après réduction éventuelle par la régulation automatique
	m_wagons = 0;		// nombre de wagons/voitures/autres locos
	m_manoeuvre = false;   // true si vitesse de manoeuvre active
	m_marche_a_vue = false; // true si le train est en marche à vue (outrepassement des signaux)
}

Train::Train(unsigned int adresse, byte pasDCC, byte vmanoeuvre,
	byte pas15, byte pas30, byte pas60, byte pas90, byte pas100, byte pas120, byte pas160, byte pas220, byte pas300, byte pas320, byte pasMax)
{
	m_adresse = adresse;
	m_pasMax = pasMax;  // pas DCC de vitesse maxi
	m_pas320 = pas320;	// pas DCC 320 km/h
	m_pas300 = pas300;
	m_pas220 = pas220;
	m_pas160 = pas160;  // pas DCC 160 km/h
	m_pas120 = pas120;
	m_pas100 = pas100;
	m_pas90 = pas90;   // pas DCC 90 km/h
	m_pas60 = pas60;   // pas DCC 60 km/h
	m_pas30 = pas30;   // pas DCC 30 km/h
	m_pas15 = pas15;   // pas DCC 30 km/h
	m_pasDCC = 28;
	m_vmanoeuvre = 0;  // numéro de fonction de la vitesse de manoeuvre
}

Train::Train(byte wagons)
{
	m_wagons = wagons;
}


Train::~Train()
{
}

void Train::ChangerSens(byte sens)
{
	if (sens == '+')
		m_sensActuel = m_sensAvant;
	else
		m_sensActuel = m_sensAvant == SENS_PAIR ? SENS_IMPAIR : SENS_PAIR;
}

void Train::RegulerVitesse(byte pas)
{
	byte train = this - Block::m_Trains;

	if (train > DERNIER_ENGIN)
		return;

	if (m_pasActuel == pas)
		return;

	TRACE_TRACKING("Train::RegulerVitesse train %u, pas de vitesse %u\n", train, pas);
#if USE_TRACER == 1
	*status_tx_old = '\0';
#endif
	C_slot_ch* pslot = GetSlot(m_adresse);

	if (pslot)
		pslot->vit = pas;

	m_pasActuel = pas;
}

void Train::Ralentir(Cote* pCote)
{
	TRACE_TRACKING("Train::Ralentir train %u\n", this - Block::m_Trains);

	if (pCote->m_vMax == VM15 || Block::Signal15kmh(pCote->m_signalSortie))
		RegulerVitesse(min(m_pasSouris, m_pas15));
	else if (pCote->m_vMax == VM30 || Block::EtatRalentissement(pCote->m_signalSortie) == RAPPEL_30)
		RegulerVitesse(min(m_pasSouris, m_pas30));
	else if (pCote->m_vMax == VM60 || Block::SignalFerme(pCote->m_signalSortie) || Block::EtatRalentissement(pCote->m_signalSortie) == RAPPEL_60)
		RegulerVitesse(min(m_pasSouris, m_pas60));
	else if (pCote->m_vMax == VM90 || Block::SignalAvertissement(pCote->m_signalSortie))
		RegulerVitesse(min(m_pasSouris, m_pas90));
	else if (pCote->m_vMax == VM100)
		RegulerVitesse(min(m_pasSouris, m_pas100));
	else if (pCote->m_vMax == VM120)
		RegulerVitesse(min(m_pasSouris, m_pas120));
	else if (pCote->m_vMax == VM160 || Block::Signal160kmh(pCote->m_signalSortie))
		RegulerVitesse(min(m_pasSouris, m_pas160));
	else if (pCote->m_vMax == VM220)
		RegulerVitesse(min(m_pasSouris, m_pas220));
	else if (pCote->m_vMax == VM300)
		RegulerVitesse(min(m_pasSouris, m_pas300));
	else if (pCote->m_vMax == VM320)
		RegulerVitesse(min(m_pasSouris, m_pas320));
	else
		RegulerVitesse(m_pasSouris);
}

void Train::ArreterOuRalentir(Cote* pCote)
{
	TRACE_TRACKING("Train::ArreterOuRalentir train %u\n", this - Block::m_Trains);

	if (Block::SignalFerme(pCote->m_signalSortie))
		RegulerVitesse(0);
	else
		Ralentir(pCote);
}

#endif
