//----------------------------------------------------------------------------------
// Module D17 par Jean-Michel Dubois
// Zone management
//----------------------------------------------------------------------------------

#include "config.h"

#if PROTOCOLE == D17PPV3

#include "Zone.h"
#include "block.h"
#include "Train.h"
#include "dcc.h"
#include "inputs.h"
#include "slot.h"

// côté d'une zone, pair ou impair

Cote::Cote()
{
	m_zoneSortie = 0;
	m_signalSortie = 0;
	m_vMax = VM320;
	m_train = LIBRE;
}

Cote::Cote(byte zoneSortie, byte signalSortie, VMax vMax)
{
	m_zoneSortie = zoneSortie;
	m_signalSortie = signalSortie;
	m_vMax = vMax;
}


Zone::Zone()
{
	m_typeZoneSensPair = ZONE_RALENTI;
	m_typeZoneSensImpair = ZONE_RALENTI;
	m_dernier_train = LIBRE;
}

Zone::Zone(
	typeZone typeZoneSensPair,
	byte zoneSortiePaire,
	byte signalSortiePair,
	VMax vMaxPair,
	typeZone typeZoneSensImpair,
	byte zoneSortieImpaire,
	byte signalSortieImpair,
	VMax vMaxImpair)
{
	m_cotePair = Cote(zoneSortiePaire, signalSortiePair, vMaxPair);
	m_typeZoneSensPair = typeZoneSensPair;
	m_coteImpair = Cote(zoneSortieImpaire, signalSortieImpair, vMaxImpair);
	m_typeZoneSensImpair = typeZoneSensImpair;
	m_dernier_train = LIBRE;

	if (m_typeZoneSensPair != m_typeZoneSensImpair && m_typeZoneSensPair != ZONE_ARRET && m_typeZoneSensPair != ZONE_RALENTI)
		TRACER.printf("ERREUR : le type d'une zone n'est pas homogène");

	for (byte i = 0; i < MAX_WIFI_CLIENTS; ++i)
		m_fModifiee[i] = false;
}

void Zone::Modifier(void)
{
	for (byte i = 0; i < MAX_WIFI_CLIENTS; ++i)
		m_fModifiee[i] = true;
}

// Renseigne les paramètres des trains présents dans une zone.

void Zone::InitTrains(byte train, byte wagons, byte sens, byte train_impair, byte wagons_impair, byte sens_impair)
{
	TRACE_TRACKING("\tZone::Init %d, train %d, wagons %d, sens %c, impair train %d, wagons %d, sens %c\n", this - Block::m_Zones, train, wagons, sens, train_impair, wagons_impair, sens_impair);

	m_dernier_train = m_cotePair.m_train = train;
	Train* pTrain = &Block::m_Trains[train];
	pTrain->m_zone = this - Block::m_Zones;
	pTrain->m_wagons = wagons;

	if (!train) {
		if (m_cotePair.m_train)
			m_cotePair.m_train = 0;
	} else if (train <= DERNIER_ENGIN) {
		pTrain->m_sensActuel = pTrain->m_sensAvant = sens == 'o' ? SENS_IMPAIR : SENS_PAIR;

		C_slot_ch* pslot = GetSlot(pTrain->m_adresse);

		// passe en marche avant si nécessaire
		if (pslot && !pslot->sens)
			pslot->sens = 1;
	} else
		pTrain->m_sensActuel = pTrain->m_sensAvant = SANS_SENS;

	m_coteImpair.m_train = train_impair;
	pTrain = &Block::m_Trains[train_impair];
	pTrain->m_zone = this - Block::m_Zones;
	pTrain->m_wagons = wagons_impair;

	if (train_impair > 127)
		pTrain->m_sensActuel = pTrain->m_sensAvant = SANS_SENS;
	else if (train) {
		pTrain->m_sensActuel = pTrain->m_sensAvant = sens == 'o' ? SENS_IMPAIR : SENS_PAIR;

		C_slot_ch* pslot = GetSlot(pTrain->m_adresse);

		// passe en marche avant si nécessaire
		if (pslot && !pslot->sens)
			pslot->sens = 1;
	}
}

void Zone::SplitTrain(byte coupure_avant, byte train_pair, byte sens_pair, byte train_impair, byte sens_impair)
{
	TRACE_TRACKING("\tZone::SplitTrain zone %d, avant %d, pair %c, sens %c, impair %d, sens %c\n", this - Block::m_Zones, coupure_avant, sens_pair, train_impair, sens_impair);

	Train* pTrainPair = &Block::m_Trains[train_pair];
	Train* pTrainImpair = &Block::m_Trains[train_impair];

	pTrainPair->m_zone = pTrainImpair->m_zone = this - Block::m_Zones;
	pTrainPair->m_wagons = coupure_avant - 1;
	pTrainImpair->m_wagons = coupure_avant;
	pTrainPair->m_sensAvant = sens_pair;
	pTrainImpair->m_sensAvant = sens_impair;
	m_cotePair.m_train = train_pair;
	m_coteImpair.m_train = train_impair;
	m_dernier_train = m_cotePair.m_train;
}

void Zone::JoinTrains(byte train, byte sens)
{
	TRACE_TRACKING("\tZone::JoinTrain zone %d, train %d, sens %c, pair %d, impair %d\n", this - Block::m_Zones, train, sens, m_cotePair.m_train, m_coteImpair.m_train);

	Train* pTrain = &Block::m_Trains[train];
	Train* pTrainPair = &Block::m_Trains[m_cotePair.m_train];
	Train* pTrainImpair = &Block::m_Trains[m_coteImpair.m_train];

	if (train == m_cotePair.m_train) {
		pTrainPair->m_wagons += pTrainImpair->m_wagons;

		if (m_coteImpair.m_train <= DERNIER_ENGIN)
			++pTrainPair->m_wagons;

		m_coteImpair.m_train = 0;
	} else {
		pTrainImpair->m_wagons += pTrainPair->m_wagons;

		if (m_cotePair.m_train <= DERNIER_ENGIN)
			++pTrainImpair->m_wagons;

		m_cotePair.m_train = m_coteImpair.m_train;
		m_coteImpair.m_train = 0;
	}

	if (train <= DERNIER_ENGIN) {
		pTrain->m_sensActuel = pTrain->m_sensAvant = sens == 'o' ? SENS_IMPAIR : SENS_PAIR;

		C_slot_ch* pslot = GetSlot(pTrain->m_adresse);

		// passe en marche avant si nécessaire
		if (pslot && !pslot->sens)
			pslot->sens = 1;
	} else
		pTrain->m_sensActuel = pTrain->m_sensAvant = SANS_SENS;

	m_dernier_train = m_cotePair.m_train;
}

// Détermine l'adresse DCC du train entrant dans la zone.
// Regarde dans la zone impaire voisine si un train roulant dans le sens pair existe, puis la zone paire voisine si un train roulant dans le sens impair existe ou existait juste avant.
// Block::UpdateZones traite d'abord les zones à train entrant puis les zones à train sortant ce qui permet à une détection ponctuelle (code à barres, rfid, cellule, etc.) de fonctionner.

void Zone::InsereTrain(void)
{
	TRACE_TRACKING("\tZone::InsereTrain dernier train connu %d\n", m_dernier_train);
	TRACE_TRACKING("\tZone::InsereTrain zone %d, sortie paire %d, sortie impaire %d\n", this - Block::m_Zones, m_cotePair.m_zoneSortie, m_coteImpair.m_zoneSortie);
	// Regarde dans la zone impaire voisine si un train existe du côté pair ou qu'il n'y a qu'un train dans la zone.
	byte train_impair = Block::m_Zones[m_coteImpair.m_zoneSortie].m_cotePair.m_train;
#if USE_TRACER == 1
	if (train_impair)
		TRACE_TRACKING("\tZone::InsereTrain essaie train impair %d, sens %c, vitesse %d, provenance zone %d\n", 
			train_impair, Block::m_Trains[train_impair].m_sensActuel, Block::m_Trains[train_impair].m_pasActuel, m_coteImpair.m_zoneSortie);
#endif
	// Le train roule-t-il en sens pair ?
	if (train_impair && train_impair <= DERNIER_ENGIN && Block::m_Trains[train_impair].m_sensActuel == SENS_PAIR) {
		// Y-a-t-il déjà un train dans la zone ?
		if (m_cotePair.m_train && m_cotePair.m_train != train_impair) {
			m_coteImpair.m_train = m_dernier_train = train_impair;	// Oui, place le train entrant côté impair.
			TRACE_TRACKING("\tZone::InsereTrain zone %d, train %d placé côté impair\n", this - Block::m_Zones, train_impair);
		} else {
			m_cotePair.m_train = train_impair;	// Non, place le train entrant du côté pair.
			TRACE_TRACKING("\tZone::InsereTrain zone %d, train %d placé côté pair\n", this - Block::m_Zones, train_impair);
		}
		// Met à jour la position de la tête de train
		Block::m_Trains[train_impair].m_zone = this - Block::m_Zones;
		return;
	}

	// Regarde dans la zone paire voisine si un train existe du côté impair.
	byte train_pair = Block::m_Zones[m_cotePair.m_zoneSortie].m_coteImpair.m_train;

	// S'il n'y a pas de train côté impair, regarde côté pair ou unique. 
	if (! train_pair)
		train_pair = Block::m_Zones[m_cotePair.m_zoneSortie].m_cotePair.m_train;
#if USE_TRACER == 1
	if (train_pair)
		TRACE_TRACKING("\tZone::InsereTrain essaie train pair %d, sens %c, vitesse %d, provenance zone %d\n", 
			train_pair, Block::m_Trains[train_pair].m_sensActuel, Block::m_Trains[train_pair].m_pasActuel, m_cotePair.m_zoneSortie);
#endif
	// Le train roule-t-il en sens impair ?
	if (train_pair && train_pair <= DERNIER_ENGIN && Block::m_Trains[train_pair].m_sensActuel == SENS_IMPAIR) {
		// Y-a-t-il déjà un train dans la zone ?
		if (m_cotePair.m_train && m_cotePair.m_train != train_pair) {
			m_coteImpair.m_train = m_dernier_train = m_cotePair.m_train;	// Oui, place le train déjà présent côté impair.
			TRACE_TRACKING("\tZone::InsereTrain zone %d, train %d déplacé côté impair\n", this - Block::m_Zones, m_coteImpair.m_train);
		}

		m_cotePair.m_train = m_dernier_train = train_pair;				// Place le train entrant du côté pair.
		TRACE_TRACKING("\tZone::InsereTrain zone %d, train %d placé côté impair\n", this - Block::m_Zones, train_pair);
		// Met à jour la position de la tête de train
		Block::m_Trains[train_pair].m_zone = this - Block::m_Zones;
		return;
	}

	if (!m_dernier_train)
		m_cotePair.m_train = m_dernier_train = INCONNU;
	else
		m_cotePair.m_train = m_dernier_train;

	TRACE_TRACKING("\tZone::InsereTrain : train perdu %d\n", m_dernier_train);
}

bool Zone::EstOccupee(void)
{
	return user_get_in(this - Block::m_Zones);
}

bool Zone::EtaitOccupee(void)
{
	return user_get_in_old(this - Block::m_Zones);
}

#endif
