//----------------------------------------------------------------------------------
// Module D17 par Jean-Michel Dubois
// Zone management
//----------------------------------------------------------------------------------

#pragma once

#include "config.h"

#if PROTOCOLE == D17PPV3

#include "inputs.h"

// Types de zones.

typedef enum typeZone {
	ZONE_RALENTI,
	ZONE_ARRET,
	ZONE_AIGUILLES,
	ZONE_MANOEUVRE,
} typeZone;

// Vitesses autorise dans les zones.

typedef enum VMax {
	VM15 = 15,
	VM30 = 30,
	VM60 = 60,
	VM90 = 90,
	VM100 = 100,
	VM120 = 120,
	VM160 = 160,
	VM220 = 220,
	VM300 = 300,
	VM320 = 320
} VMax;

// Position des trains dans la zone, train Unique, train ct pair (E), train ct impair (O)

enum { 
	UNIQUE = 'U', 
	COTE_PAIR = 'E', 
	COTE_IMPAIR = 'O' 
};

constexpr auto LIBRE = 0;
constexpr auto HEURTOIR = 255;

// Cot d'une zone.

class Cote
{
public:
	Cote(byte zoneSortie, byte signalSortie, VMax vMax);
	Cote();
	
	byte m_zoneSortie;		// numro de la zone de sortie
	byte m_signalSortie;	// numro du signal protgeant l'accs de la zone de sortie
	VMax m_vMax;			// vitesse maxi
	byte m_train;			// ID de train ou 0 = LIBRE
};

class Zone
{
public:
	Zone();
	Zone(
		typeZone typeZoneSensPair,
		byte zoneSortiePaire,
		byte signalSortiePair,
		VMax vMaxPair,
		typeZone typeZoneSensImpaire,
		byte zoneSortieImpaire,
		byte signalSortieImpaire,
		VMax vMaxImpair);

	void InitTrains(byte train, byte wagons, byte sens, byte m_train_impair, byte wagons_impair, byte sens_impair);
	void SplitTrain(byte coupure_avant, byte train_pair, byte sens_pair, byte train_impair, byte sens_impair);
	void JoinTrains(byte nouveau_train, byte sens);
	void InsereTrain(void);
	bool EstOccupee(void);
	bool EtaitOccupee(void);
	inline bool EstLibre(void) { return !EstOccupee(); }
	inline bool EtaitLibre(void) { return !EtaitOccupee(); }
	void Modifier(void);				// Marquer la zone comme modifie pour tous les clients

	Cote		m_cotePair;					// Paramtres du ct pair de la zone
	typeZone	m_typeZoneSensPair;			// manoeuvre, ralentissement, arrt, aiguilles
	Cote		m_coteImpair;				// Paramtres du ct impair de la zone
	typeZone	m_typeZoneSensImpair;		// manoeuvre, ralentissement, arrt, aiguilles
	short		m_dernier_train;			// adresse DCC, ID de train passif ou 0 = LIBRE du dernier train ayant occup la zone.
	bool		m_fModifiee[MAX_WIFI_CLIENTS];	// true si des informations de la zone ont t modifies depuis la dernire demande d'tat par chaque client
};

#endif