#pragma once

#include <ESP8266WiFi.h>
#include <stdarg.h>
#include "defines.h"

// Protocoles disponibles
#define D17				0		// D17 d'origine : TCO simple
#define D17PPV2			1		// D17 + vitesse réelle des trains, réinitialisation à distance
#define D17PPV3			2		// D17PPV2 + TCO avec suivi des trains, jonction, coupure, mesure de vitesse, reset
#define DCCPP			3		// DCC++ avec cordon FTDI (en version simplifiee sans configuration, mais ok pour Rocrail)
#define CDMRAIL			4		// DCC++ avec cordon FTDI (en version simplifiee sans configuration, mais ok pour CDM-Rail)
#define DSTATION		5		// Desktop Station

//--------------------------------------------------------------------------------
// Constantes utilisateur (vous pouvez changer ces valeurs)
//----------------------------------------------------------------------------------

// 1) Protocole utilisé
#define PROTOCOLE		D17PPV3

#if PROTOCOLE == D17PPV3
// Block systems parameters
constexpr auto DERNIER_TRAIN = 10;		// Identifiant de train avec machine le plus élevé utilisé. 127 maxi. Utilisé uniquement pour raccourcir des traitements.
constexpr auto NB_AIGUILLES = 1;		// Nombre d'aiguilles

extern bool fUseBlock;					// true si au moins un client supporte le suivi des trains
#endif

/* 2) configuration du WIFI */

#define USE_WIFI	1			// 1 = la communication entre la centrale et le logiciel de conduite est en WiFi, 0 = la communication est en série

// Mode
#define APMODE 0				// 0 = wifi station, wifi station (D17 need to be connected to an external WIFI AP)

#if APMODE
# define DEFAULT_AP_CONFIG 1	// 1 = default access point network configuration (IP 192.168.4.1), 0 = custom network configuration

constexpr auto WIFI_SSID = "D17-0001";
constexpr auto WIFI_PASS = "ulysse31";
#else
// wifi station, use custom configuration
constexpr auto WIFI_SSID = "Livebox-404A-EXT";
constexpr auto WIFI_PASS = "4pejNGUfW5CJpty7TU";
#endif

// custom network configuration
#if ! APMODE || ! DEFAULT_AP_CONFIG
# define IPADDRESS (192, 168, 1, 251)
# define GATEWAY   (192, 168, 1, 1)
# define SUBNET    (255, 255, 255, 0)
#endif

// Server configuration
#define TCP_PORT 1234

// Max parallel wifi connections (can be from different devices)
constexpr auto MAX_WIFI_CLIENTS = 4;

// Nombre de souris par dispositif de commande
constexpr auto MAX_CHANNELS = 4;

// Forcage de l'utilisation de l'USB
#if PROTOCOLE > D17PPV3 && USE_WIFI
#undef USE_WIFI
#define USE_WIFI 0
#endif

/* 4) DCC configuration */

// Slot number used to store loco commands and send them in loop.
#define NB_SLOTS 12      // 1 min, 12 by default

// F0-F28 packets are repeated FCT_AUX_REPEAT times, minimal value 1, 255 = repeat forever
constexpr auto FCT_AUX_REPEAT = 3;

// CV programmation code
constexpr auto CV_PROG_CODE = 1234;

/* 5) Configuration des bus, voyants et boutons, tous protocoles */

// PIN assignment
#define USER_USE_S88  1     // 1 to use S88 IN     (out D0, D5, D1, D8 no more available)
#define USER_USE_MAX  1     // 1 to use MAX7219/21 (out D0, D5, D4     no more available)
#define USER_USE_PCA  1		// 1 to use PCA9685    (out D6, D7         no more available)
#define USER_USE_MCPIN  0   // 1 to use MCP23017   (out D6, D7         no more available)
#define USER_USE_MCPOUT 0   // 1 to use MCP23017   (out D6, D7         no more available)
#define USER_CLI_LED_MAX -1 // indicate which LED of MAX should blink to indicate sign of life (-1 not used) 
#define USER_USE_AU   1     // 1 to use AU button on pin D5 (it can be used in // with S88, MAX, OUT) (very small glitch with OUT). Read the doc because 2 resistors must be used !
#define USER_AU_LEVEL 0     // 0 AU actif a l'etat bas, 1 AU actif a l'etat haut

#if PROTOCOLE == D17PPV3
#define USER_USE_D19	1	// 1 to use D19 light signal module
#else
#define USER_USE_D19	0
#endif

#if USER_USE_S88
# define USER_S88_MODULES			1	// nombre de modules S88 (1 <= n <= 6)
#else
# define USER_S88_MODULES			0
#endif
#if USER_USE_MCPIN
# define USER_MCP23017_IN_MODULES	1	// nombre de modules MCP23017 en entr�e (1 <= n <= 6)
#else
# define USER_MCP23017_IN_MODULES	0
#endif
#if USER_USE_MCPOUT
// nombre de modules MCP23017 en sortie (1 <= n <= 8)
// le nombre maximal de modules MCP23017 est 8
// l'adresse I2C du premier module d'entr�e doit �tre sup�rieure ou �gale � celle du dernier module de sortie + 1
// le num�ro des 16 sorties du premier module est compris entre 16 et 31, du deuxi�me entre 32 et 47 et ainsi de suite
# define USER_MCP23017_OUT_MODULES	1
#else
# define USER_MCP23017_OUT_MODULES	0
#endif

// speed mode
#define USER_SPEED_STEPS 28   // 28 crans ou 126 crans
#define USER_PWM_FACTOR 100   // Plage d'utilisation de la pwm pour les locos analogiques (en %) 
                              // afin de reduire la tension moyenne max si si l'alim > 12V. 
                              // (ex: 12V * 100% = 12V, 15V * 80% = 12V, 18V * 67% = 12V)

// dcc & analog modes                              
#define USER_DCC_OK  1        // 0 ne jamais utiliser le DCC, 1 ok pour utiliser le DCC
#define USER_ANA_OK  0        // 0 ne jamais utiliser l'analogique, 1 ok pour utilise l'analogique   
#define USER_ACC_ANA 0        // commande des accessoires en analogique: 0 ne pas commander, 1 commander sans le booster, 2 commander avec le booster                      

// parametres pour le protocol DCC++     

#if PROTOCOLE == DCCPP
constexpr auto DCCPP_SENSOR_NB = 16;    // number of IN to use with DCC++ (96 max) Need to be a multiple of 8
#endif

/* 6) Configuration du mode trace pour la mise au point */

// Le mode trace WiFi envoie les informations sur un port TCP. On peut les voir avec un client telnet.

#define USE_TRACER			1	// 0 = pas de trace, 1 = trace USB, 2 = trace WiFi

#if USE_TRACER == 0
# define TRACER Serial
# define TRACE1(...)
#elif USE_TRACER == 1
# define TRACER Serial
# define TRACE1(...) TRACER.printf(__VA_ARGS__)
#else
# define TRACER trclient
# define TRACE1(...) if (TRACER) TRACER.printf(__VA_ARGS__), TRACER.print("\r")
# define TRACER_PORT	23			// Port par defaut du serveur de trace
extern WiFiClient trclient;
#endif

#if USE_TRACER
# define PRINT_PACKET       0     // affichage des paquets de bits DCC
# define PRINT_TIME         0     // affichage du temps de boucle
# define PRINT_CAN          0     // affichage de la mesure de courant
# define PRINT_STATUS		1	  // affichage de la ligne de statut
# define PRINT_COMMAND		1	  // affichage de la ligne de commande
# define PRINT_SIGNAL		1     // affichage des changement des signaux
# define PRINT_TRACKING		1	  // affichage des informations de suivi des trains
# define PRINT_DECODER		1	  // affichage de la programmation des decodeurs

# if PRINT_PACKET
// types of packets printed by dcc_tx2 and dcc_tx3
extern byte print_idle_packet;
extern byte print_speed_packet;
extern byte print_function_packet;
extern byte print_accessory_packet;
extern byte print_other_packet;
extern bool speed_packet;
# endif
#endif

#if PRINT_TRACKING
# define TRACE_TRACKING(...) TRACE1(__VA_ARGS__)
#else
# define TRACE_TRACKING(...)
#endif

#if PRINT_DECODER
# define TRACE_DECODER(...) TRACE1(__VA_ARGS__)
#else
# define TRACE_DECODER(...)
#endif

#if PROTOCOLE == D17PPV3
#if ! USER_USE_S88 && ! USER_USE_MCPIN
#error Train tracking requires USER_USE_S88 1 or USER_USE_MCPIN 1
#endif
#if USER_USE_D19 && ! USER_USE_MAX
#error D19 modules require USER_USE_MAX 1
#endif
#endif

// Version de la centrale
constexpr auto CENTRAL_NAME = "D17++ v3-20191004";

extern char status_tx[256];
extern char tx[256];
#if USE_TRACER == 1
extern char status_tx_old[256];
#endif
