#include "d19.h"

#if USER_USE_D19 && USER_USE_MAX

#include "signal.h"
#include "leds.h"

d19::d19()
{
}


d19::~d19()
{
}

// Interface SPI

// position des LED dans la matrice

// 0 = segg
// 1 = segf
// 2 = sege
// 3 = segd
// 4 = segc
// 5 = segb
// 6 = sega
// 7 = segdp

// 0 = dig0
// 8 = dig1
// 16 = dig2
// 24 = dig3
// 32 = dig4
// 40 = dig5
// 48 = dig6
// 56 = dig7

#if ! defined(nobreak)
# define nobreak
#endif

void d19::a2feux(Signal* pSignal)
{
	// LED de base des signaux du module 32 signaux  deux feux

	static byte led[] = {
	   6, 14, 22, 30, 38, 46, 54, 62,
	   4, 12, 20, 28, 36, 44, 52, 60,
	   2, 10, 20, 26, 34, 42, 50, 58,
	   0,  8, 16, 24, 32, 40, 48, 56
	};

	enum {
		_2FEUX_B = 0, _2FEUX_VI
	};

	if (pSignal->m_uiAdresse >= IO_LED_NB / 64 || pSignal->m_bSignal >= sizeof(led))
		return;

	byte base = pSignal->m_uiAdresse * 64 + led[pSignal->m_bSignal];

	byte b = base + _2FEUX_B;
	byte vi = base + _2FEUX_VI;

	user_led(b, 0);
	user_led(vi, 0);
	user_led_cli(b, 0);

	switch (pSignal->m_Feux) {
	case MANOEUVRE_CLIGNO:
		user_led_cli(b, 1);
		nobreak;
	case MANOEUVRE:
		user_led(b, 1);
		break;
	case CARRE_VIOLET:
		user_led(vi, 1);
		break;
	default:
		break;
	}
}

void d19::a3feux(Signal* pSignal)
{
	// LED de base des signaux  3 feux du module 8 signaux  3 feux et 8 signaux  5 feux

	static byte led[] = {
	   0, 16, 32, 48,
	   8, 24, 40, 56
	};

	enum {
		_3FEUX_J = 0, _3FEUX_R, _3FEUX_V,
		_3FEUX_VI = 1, _3FEUX_BL
	};

	if (pSignal->m_uiAdresse >= IO_LED_NB / 64 || pSignal->m_bSignal >= sizeof(led))
		return;

	byte base = pSignal->m_uiAdresse * 64 + led[pSignal->m_bSignal];

	byte v = base + _3FEUX_V; // vert ou blanc
	byte r = base + _3FEUX_R; // rouge ou violet
	byte j = base + _3FEUX_J;

	user_led(v, 0);
	user_led(r, 0);
	user_led(j, 0);
	user_led_cli(v, 0);
	user_led_cli(r, 0);
	user_led_cli(j, 0);

	switch (pSignal->m_Feux) {
	case MANOEUVRE_CLIGNO:
	case VOIE_LIBRE_CLIGNO:
		user_led_cli(v, 1);
		nobreak;
	case MANOEUVRE:
	case VOIE_LIBRE:
		user_led(v, 1);
		break;
	case AVERTISSEMENT_CLIGNO:
		user_led_cli(j, 1);
		nobreak;
	case AVERTISSEMENT:
		user_led(j, 1);
		break;
	case SEMAPHORE_CLIGNO:
		user_led_cli(r, 1);
		nobreak;
	case CARRE_VIOLET:
	case SEMAPHORE:
		user_led(r, 1);
		break;
	default:
		break;
	}
}

void d19::a4feux(Signal* pSignal)
{
	// LED de base des signaux du module 16 signaux  4 feux

	static byte led[] = {
	   4, 20, 36, 52,
	   0, 16, 32, 48,
	  12, 28, 44, 60,
	   8, 24, 40, 56
	};

	enum {
		_4FEUX_J = 0, _4FEUX_R, _4FEUX_V, _4FEUX_RC
	};

	if (pSignal->m_uiAdresse >= IO_LED_NB / 64 || pSignal->m_bSignal >= sizeof(led))
		return;

	byte base = pSignal->m_uiAdresse * 64 + led[pSignal->m_bSignal];

	byte rc = base + _4FEUX_RC;
	byte v = base + _4FEUX_V; // vert ou blanc
	byte r = base + _4FEUX_R; // rouge ou violet
	byte j = base + _4FEUX_J;

	user_led(rc, 0);
	user_led(v, 0);
	user_led(r, 0);
	user_led(j, 0);
	user_led_cli(v, 0);
	user_led_cli(r, 0);
	user_led_cli(j, 0);

	switch (pSignal->m_Feux) {
	case MANOEUVRE_CLIGNO:
	case VOIE_LIBRE_CLIGNO:
		user_led_cli(v, 1);
		nobreak;
	case MANOEUVRE:
	case VOIE_LIBRE:
		user_led(v, 1);
		break;
	case AVERTISSEMENT_CLIGNO:
		user_led_cli(j, 1);
		nobreak;
	case AVERTISSEMENT:
		user_led(j, 1);
		break;
	case SEMAPHORE_CLIGNO:
		user_led_cli(r, 1);
		nobreak;
	case CARRE_VIOLET:
	case SEMAPHORE:
		user_led(r, 1);
		break;
	case CARRE:
		user_led(r, 1);
		user_led(rc, 1);
		break;
	default:
		break;
	}
}

void d19::a5feux(Signal* pSignal)
{
	// LED de base des signaux  5 feux du module 8 signaux  3 feux et 8 signaux  5 feux
	static byte led[] = {
	   3, 19, 35, 51,
	  11, 27, 43, 59
	};

	enum {
		_5FEUX_J = 0, _5FEUX_R, _5FEUX_V, _5FEUX_RC, _5FEUX_JR
	};

	if (pSignal->m_uiAdresse >= IO_LED_NB / 64 || pSignal->m_bSignal >= sizeof(led))
		return;

	byte base = pSignal->m_uiAdresse * 64 + led[pSignal->m_bSignal];

	byte rc = base + _5FEUX_RC;
	byte v = base + _5FEUX_V; // vert ou blanc
	byte r = base + _5FEUX_R; // rouge ou violet
	byte j = base + _5FEUX_J;
	byte jr = base + _5FEUX_JR;

	user_led(rc, 0);
	user_led(v, 0);
	user_led(r, 0);
	user_led(j, 0);
	user_led(jr, 0);
	user_led_cli(v, 0);
	user_led_cli(r, 0);
	user_led_cli(j, 0);
	user_led_cli(jr, 0);

	switch (pSignal->m_Feux) {
	case MANOEUVRE_CLIGNO:
	case VOIE_LIBRE_CLIGNO:
		user_led_cli(v, 1);
		nobreak;
	case MANOEUVRE:
	case VOIE_LIBRE:
		user_led(v, 1);
		break;
	case AVERTISSEMENT_CLIGNO:
		user_led_cli(j, 1);
		nobreak;
	case AVERTISSEMENT:
		user_led(j, 1);
		break;
	case SEMAPHORE_CLIGNO:
		user_led_cli(r, 1);
		nobreak;
	case CARRE_VIOLET:
	case SEMAPHORE:
		user_led(r, 1);
		break;
	case CARRE:
		user_led(r, 1);
		user_led(rc, 1);
		break;
	default:
		break;
	}

	switch (pSignal->m_Ralenti) {
	case RALENTISSEMENT_60:
	case RAPPEL_60:
		user_led_cli(jr, 1);
		nobreak;
	case RALENTISSEMENT_30:
	case RAPPEL_30:
		user_led(jr, 1);
		user_led(jr, 1);
		break;
	default:
		break;
	}
}

// 6 feux en utilisant deux sorties  3 feux d'un module  8x5 feux + 8x3 feux

// feu 1 : jaune avertissement, rouge, vert
// feu 2 : blanc, rouge carr, jaune ralentissement x 2
// Le numro de signal doit tre compris entre 0 et 6.
// Le signal utilise la sortie  3 feux indique dans la table des signaux et la suivante.

void d19::a2x3feux(Signal* pSignal)
{
	// LED de base des signaux  3 feux du module 8 signaux  3 feux et 8 signaux  5 feux

	static byte led[] = {
	   0, 16, 32, 48,
	   8, 24, 40, 56
	};

	enum {
		_3FEUX_J = 0, _3FEUX_R, _3FEUX_V,
		_3FEUX_RC, _3FEUX_BL = 0, _3FEUX_JR
	};

	if (pSignal->m_uiAdresse >= IO_LED_NB / 64 || pSignal->m_bSignal >= sizeof(led) - 1)
		return;

	// led de base de la premiere sortie  3 feux
	byte base = pSignal->m_uiAdresse * 64 + led[pSignal->m_bSignal];

	byte v = base + _3FEUX_V; // vert ou blanc
	byte r = base + _3FEUX_R; // rouge ou violet
	byte j = base + _3FEUX_J;
	// led de base de la deuxime sortie  3 feux
	base = pSignal->m_uiAdresse * 64 + led[pSignal->m_bSignal + 1];

	byte bl = base + _3FEUX_BL;
	byte rc = base + _3FEUX_RC;
	byte jr = base + _3FEUX_JR;

	user_led(v, 0);
	user_led(r, 0);
	user_led(j, 0);
	user_led(bl, 0);
	user_led(rc, 0);
	user_led(jr, 0);
	user_led_cli(v, 0);
	user_led_cli(r, 0);
	user_led_cli(j, 0);
	user_led_cli(bl, 0);
	user_led_cli(jr, 0);

	switch (pSignal->m_Feux) {
	case MANOEUVRE_CLIGNO:
		user_led_cli(bl, 1);
		nobreak;
	case MANOEUVRE:
		user_led(bl, 1);
		break;
	case VOIE_LIBRE_CLIGNO:
		user_led_cli(v, 1);
		nobreak;
	case VOIE_LIBRE:
		user_led(v, 1);
		break;
	case AVERTISSEMENT_CLIGNO:
		user_led_cli(j, 1);
		nobreak;
	case AVERTISSEMENT:
		user_led(j, 1);
		break;
	case SEMAPHORE_CLIGNO:
		user_led_cli(r, 1);
		nobreak;
	case CARRE_VIOLET:
	case SEMAPHORE:
		user_led(r, 1);
		break;
	case CARRE:
		user_led(r, 1);
		user_led(rc, 1);
		break;
	default:
		break;
	}
}

#undef nobreak

#endif