//--------------------------------------------------------------------------------
// Module D17 par Ulysse Delmas-Bègue et Jean-Michel Dubois
// Génération du signal DCC
//--------------------------------------------------------------------------------

#include "dcc.h"
#include "slot.h"

//--------------------------------------------------------------------------------
// Variables
//----------------------------------------------------------------------------------

byte boost = BOOST_DCC;
byte oldboost = BOOST_DCC;
byte au = 1;   //maintenant la centrale demarre en AU (donc alimenation off, cela est plus sure)  
byte pwmsens = 0;
unsigned int pwm1024 = 0;
unsigned int cv_adr = 1;
byte cv_dat = 3;
unsigned int an;   //derniere lecture du convertisseur analogique/numerique

#if PRINT_PACKET
// types of packets printed by dcc_tx2 and dcc_tx3
byte print_idle_packet = 0;
byte print_speed_packet = 1;
byte print_function_packet = 1;
byte print_accessory_packet = 1;
byte print_other_packet = 1;
bool speed_packet = false;
#endif

//--------------------------------------------------------------------------------
// DCC signal   part
//----------------------------------------------------------------------------------

// DCC
// "1"
//    typ      58us
//    station  55-61us  (delta max entre periodes 3us)  --> measure LA OK: 1=57.5 or 61us
//    decodeur 52-64us  (delta max entre periodes 6us)     
// "0"
//    typ      100us
//    station  95-9990us (total 12000us)
//
//
// RESET    : 1111111111111111 0 00000000 0 00000000            0 00000000 1
// IDLE     : 1111111111111111 0 11111111 0 00000000            0 EEEEEEEE 1
//
// VIT  28  : 1111111111111111 0 0AAAAAAA 0 01DFSSSS            0 EEEEEEEE 1
// VIT 127  : 1111111111111111 0 0AAAAAAA 0 00111111            0 DSSSSSSS 0 EEEEEEEE 1
//
// FCT 0-4  : 1111111111111111 0 0AAAAAAA 0 100-FL-F4-F3-F2-F1  0 EEEEEEEE 1
// FCT 5-8  : 1111111111111111 0 0AAAAAAA 0 1011-F8-F7-F6-F5    0 EEEEEEEE 1 
// FCT 9-12 : 1111111111111111 0 0AAAAAAA 0 1010-F12-F11-F10-F9 0 EEEEEEEE 1 
// FCT 13-20: 1111111111111111 0 0AAAAAAA 0 11011110            0 F20-F19-F18-F17-F16-F15-F14-F13 0 EEEEEEEE 1 
// FCT 11-28: 1111111111111111 0 0AAAAAAA 0 11011111            0 F28-F27-F26-F25-F24-F23-F22-F21 0 EEEEEEEE 1 
//
//
// Pour les adresses longues (seulement pour les decodeurs mobiles):
// - Le champ d'adresse (0-127):  0AAAAAAA  (0 pour le broadcast)    cv1
// - est remplace par (0-10239):  11AAAAAA AAAAAAAA (0-39)           cv17 192+AAAAAA(0-39)  cv18 AAAAAAAA (0-255)
// - choix cv29.5 xx1xxxxx=long (126/28 crans independant de l'adressage court/long)
//
// ACC 1-511: 111111111111111 0 10AAAAAA 0 1AAA1DDD            0 EEEEEEEE 1 attention, les 3xAAA sont inverses, ils forment aussi le MSB adr0 est reserve le 1 entre AetD peut etre mis a 0 pour mettre la sortie a 0, mais normalement allumer la 0 eteint la 1 et inversement. ou alors remise a 0 automatique apres une pulse
// ACC1-2047: 111111111111111 0 10AAAAAA 0 0AAA0AA1 0 000xxxxx 0 EEEEEEEE 1
//
// une explication accessible sur le DCC est dispo sur: http://trainminiature.discutforum.com/t12784-dcc-comment-ca-marche-place-a-la-technique

#define T1 55   // half period DCC bit 1 
#define T0 100  // half period DCC bit 0

// jmd>>

static void dcc_byte(byte a)
{
	byte i;

	// separator
	digitalWrite(DCC_PIN, LOW); delayMicroseconds(T0); digitalWrite(DCC_PIN, HIGH); delayMicroseconds(T0); // delay(0);

	for (i = 8; i; i--)    //MSB transmit first
	{
		if (a & 128) { digitalWrite(DCC_PIN, LOW); delayMicroseconds(T1); digitalWrite(DCC_PIN, HIGH); delayMicroseconds(T1); }
		else { digitalWrite(DCC_PIN, LOW); delayMicroseconds(T0); digitalWrite(DCC_PIN, HIGH); delayMicroseconds(T0); }
		a <<= 1;
	}
}

static byte dcc_tx(byte p_len, byte nb, byte a, byte b, byte c = 0, byte d = 0)
{
	byte x, i;

	// preambule
	for (i = p_len; i; i--)
	{
		digitalWrite(DCC_PIN, LOW); delayMicroseconds(T1); digitalWrite(DCC_PIN, HIGH); delayMicroseconds(T1);
	}

	dcc_byte(x = a);

	if (nb > 1) {
		x ^= b;
		dcc_byte(b);

		if (nb > 2) {
			x ^= c;
			dcc_byte(c);

			if (nb > 3) {
				x ^= d;
				dcc_byte(d);
			}
		}
	}

	// checksum
	dcc_byte(x);

	// end
	digitalWrite(DCC_PIN, LOW); delayMicroseconds(T1); digitalWrite(DCC_PIN, HIGH); delayMicroseconds(T1);

	// start a 0 (can last 9.9ms max)
	digitalWrite(DCC_PIN, LOW); delayMicroseconds(T0); digitalWrite(DCC_PIN, HIGH); delayMicroseconds(T0); delay(0);
	return x;
}

#if PRINT_PACKET
unsigned long int boosterloop = 0;

void print_preamble(byte i)
{
	for (; i; i--)
		TRACER.print("1");
}

static void print_byte(byte a)
{
	byte i;

	TRACER.print(" 0 ");

	for (i = 8; i; i--) {
		TRACER.print(a & 128 ? "1" : "0");
		a <<= 1;
	}
}

static void print_tx2_packet(const char* s, byte p_len, byte a, byte c, byte x)
{
	TRACER.printf("loop %lu %lums %s", boosterloop, millis(), s);
	print_preamble(p_len);
	print_byte(a);
	print_byte(c);
	print_byte(x);
	TRACER.println(" 1");
}

static void print_tx3_packet(const char* s, byte p_len, byte a, byte b, byte c, byte x)
{
	TRACER.printf("loop %lu %lums %s", boosterloop, millis(), s);
	print_preamble(p_len);
	print_byte(a);
	print_byte(b);
	print_byte(c);
	print_byte(x);
	TRACER.println(" 1");
}

static void print_tx4_packet(const char* s, byte p_len, byte a, byte b, byte c, byte d, byte x)
{
	TRACER.printf("loop %lu %lums %s", boosterloop, millis(), s);
	print_preamble(p_len);
	print_byte(a);
	print_byte(b);
	print_byte(c);
	print_byte(d);
	print_byte(x);
	TRACER.println(" 1");
}

#endif  // ? PRINT_PACKET

void dcc_tx2(byte p_len, byte a, byte c)
{
#if PRINT_PACKET
	byte x;

	if (a == 0xff) {
		x = dcc_tx(p_len, 2, a, c, 0);
		speed_packet = false;

		if (print_idle_packet)
			print_tx2_packet("IDLE     ", p_len, a, c, x);
		return;
	}

	if ((c & 0xc0) == 0x40) {
		x = dcc_tx(p_len, 2, a, c, 0);

		if (print_speed_packet && !speed_packet)
			print_tx2_packet("SPEED    ", p_len, a, c, x);

		speed_packet = true;
		return;
	}

	if ((c & 0xE0) == 0x80) {
		x = dcc_tx(p_len, 2, a, c, 0);
		speed_packet = false;

		if (print_function_packet)
			print_tx2_packet("FCT0_4   ", p_len, a, c, x);

		return;
	}

	if ((c & 0xF0) == 0xB0) {
		x = dcc_tx(p_len, 2, a, c, 0);
		speed_packet = false;

		if (print_function_packet)
			print_tx2_packet("FCT5_8   ", p_len, a, c, x);

		return;
	}

	if ((c & 0xF0) == 0xA0) {
		x = dcc_tx(p_len, 2, a, c, 0);
		speed_packet = false;

		if (print_function_packet)
			print_tx2_packet("FCT9_12  ", p_len, a, c, x);

		return;
	}

	if ((a & 0xC0) == 0x80) {
		x = dcc_tx(p_len, 2, a, c, 0);
		speed_packet = false;

		if (print_accessory_packet)
			print_tx2_packet("BASICACC ", p_len, a, c, x);

		return;
	}

	if (print_other_packet) {
		x = dcc_tx(p_len, 2, a, c, 0);
		speed_packet = false;
		print_tx2_packet("???????? ", p_len, a, c, x);
		return;
	}
#endif
	dcc_tx(p_len, 2, a, c);
}


void dcc_tx3(byte p_len, byte a, byte b, byte c)
{
#if PRINT_PACKET
	byte x;

	if (b == 0b11011110) {
		x = dcc_tx(p_len, 3, a, b, c);

		if (print_function_packet)
			print_tx3_packet("FCT13_20 ", p_len, a, b, c, x);

		return;
	}

	if (b == 0b11011111) {
		x = dcc_tx(p_len, 3, a, b, c);

		if (print_function_packet)
			print_tx3_packet("FCT28_21 ", p_len, a, b, c, x);

		return;
	}

	if (print_other_packet) {
		x = dcc_tx(p_len, 3, a, b, c);
		print_tx3_packet("??????? ", p_len, a, b, c, x);
		return;
	}
#endif
	dcc_tx(p_len, 3, a, b, c);
}

void dcc_tx4(byte p_len, byte a, byte b, byte c, byte d)
{
#if PRINT_PACKET
	byte x;

	if (print_other_packet) {
		x = dcc_tx(p_len, 3, a, b, c);
		print_tx4_packet("??????? ", p_len, a, b, c, d, x);
		return;
	}
#endif
	dcc_tx(p_len, 4, a, b, c, d);
}

void dcc_keepalive(void)   //a appeller si une fonction dure plus de 9.9ms pour que les decodeurs restent en mode DCC
{
	if (boost != BOOST_DCC) return;

	// start a 0 (can last 9.9ms max)
	digitalWrite(DCC_PIN, LOW); delayMicroseconds(T0); digitalWrite(DCC_PIN, HIGH); delayMicroseconds(T0); delay(0);
}


// envoie d'une trame pour activer une sortie sur un decodeur d'accessoire basique
// normalement val=1 car les decodeurs d'accesoires sont souvent configures soit:
// - allumer la 0 eteint la 1 et vice versa
// - generer des pulses
// adr: 1-511
// out: 0-7
// val: 0/1
// ACC1-511: 1111111111111111 0 10AAAAAA 0 1AAAVDDD 0 EEEEEEEE 1
void user_bas_acc_dec_tx(unsigned int adr, byte out, byte val)
{
	byte b1 = 128 | ((byte)adr & 0x3f);
	byte b2 = (((byte)(adr >> 2) & 0x70) ^ 0x70) | 0x80 | (out & 7) | ((val & 1) << 3);

	if (boost == BOOST_ANA)
	{
#if USER_ACC_ANA == 0   // pas de transmission en ANA 
		return;
#endif
#if USER_ACC_ANA == 1   // desactivation du booster durant la transmission
		if (au == 0) analogWrite(P_PIN, 0);
#endif
#if USER_ACC_ANA == 2   // activation du booster durant la transmission
		if (au == 0) analogWrite(P_PIN, 1);
#endif        
	}

	dcc_tx2(16, b1, b2);  //on transmet 2x une commande pour un accessoire
	dcc_tx2(16, b1, b2);

	if (boost == BOOST_ANA)
	{
		if (pwmsens) digitalWrite(S_PIN, HIGH);
		else        digitalWrite(S_PIN, LOW);

#if USER_ACC_ANA == 1   // desactivation du booster durant la transmission
		if (au == 0) analogWrite(P_PIN, pwm1024);
#endif
#if USER_ACC_ANA == 2   // activation du booster durant la transmission
		if (au == 0) analogWrite(P_PIN, pwm1024);
#endif        
	}
}


// envoie d'une valeur 0-31 a un decodeur d'accessoire etendu
// adr: 1-2044
// val: 031
// ACC1-2044: 111111111111111 0 10AAAAAA 0 0AAA0AA1 0 000DDDDD 0 EEEEEEEE 1
void user_ext_acc_dec_tx(unsigned int adr, byte val)
{
	adr--; // 0-2043
	byte low = adr & 3; //0-3 
	adr = adr >> 2; //0-510
	adr++; //1-511

	byte b1 = 128 | (byte)(adr & 0x3f);
	byte b2 = (((byte)(adr >> 2) & 0x70) ^ 0x70) | (low << 1) | 1;

	if (boost == BOOST_ANA)
	{
#if USER_ACC_ANA == 0   // pas de transmission en ANA 
		return;
#endif
#if USER_ACC_ANA == 1   // desactivation du booster durant la transmission
		if (au == 0) analogWrite(P_PIN, 0);
#endif
#if USER_ACC_ANA == 2   // activation du booster durant la transmission
		if (au == 0) analogWrite(P_PIN, 1);
#endif        
	}

	dcc_tx3(16, b1, b2, val);  //on transmet 2x une commande pour un accesoire
	dcc_tx3(16, b1, b2, val);

	if (boost == BOOST_ANA)
	{
		digitalWrite(S_PIN, pwmsens ? HIGH : LOW);

#if USER_ACC_ANA == 1   // desactivation du booster durant la transmission
		if (au == 0) analogWrite(P_PIN, pwm1024);
#endif
#if USER_ACC_ANA == 2   // activation du booster durant la transmission
		if (au == 0) analogWrite(P_PIN, pwm1024);
#endif        
	}
}


// sens: 1=AV, 0=AR
// vit 0-28
static byte vit28_2_dcc(byte sens, byte vit)
{
	if (vit > 0)
	{
		vit += 3;
		if (vit & 1) vit |= 32;
		vit = vit >> 1;
	}
	if (sens) vit |= 32;
	vit |= 64;

	return vit;
}
// vit 0-126
static byte vit126_2_dcc(byte sens, byte vit)
{
	if (vit > 0) vit += 1;
	if (sens) vit |= 128;

	return vit;
}



//--------------------------------------------------------------------------------
// DCC prog   part
//----------------------------------------------------------------------------------

enum {
	DCC_PROG_FRAME_IDLE,
	DCC_PROG_FRAME_RESET,
	DCC_PROG_FRAME_PAGE,
	DCC_PROG_FRAME_DAT,
	DCC_PROG_FRAME_VER,
	DCC_PROG_FRAME_ACK,
	DCC_PROG_FRAME_END
};

static byte dcc_write_seq[] = { 
	DCC_PROG_FRAME_RESET, 
	DCC_PROG_FRAME_RESET, 
	DCC_PROG_FRAME_RESET, 

	DCC_PROG_FRAME_PAGE, 
	DCC_PROG_FRAME_PAGE, 
	DCC_PROG_FRAME_PAGE, 
	DCC_PROG_FRAME_PAGE, 
	DCC_PROG_FRAME_PAGE, 

	DCC_PROG_FRAME_RESET, 
	DCC_PROG_FRAME_RESET, 
	DCC_PROG_FRAME_RESET, 
	DCC_PROG_FRAME_RESET, 
	DCC_PROG_FRAME_RESET, 
	DCC_PROG_FRAME_RESET, 
	DCC_PROG_FRAME_RESET, 
	DCC_PROG_FRAME_RESET, 
	DCC_PROG_FRAME_RESET, 

	DCC_PROG_FRAME_DAT, 
	DCC_PROG_FRAME_DAT, 
	DCC_PROG_FRAME_DAT, 
	DCC_PROG_FRAME_DAT, 
	DCC_PROG_FRAME_DAT, 

	DCC_PROG_FRAME_ACK, 
	DCC_PROG_FRAME_ACK, 
	DCC_PROG_FRAME_ACK, 
	DCC_PROG_FRAME_ACK, 
	DCC_PROG_FRAME_ACK, 
	DCC_PROG_FRAME_ACK, 

	DCC_PROG_FRAME_END
};

#define WRITE_SEQ_SIZE sizeof(dcc_write_seq) / sizeof(dcc_write_seq[0])

static unsigned int MeanCurrent(void)
{
	constexpr auto MEASURES = 100;
	// mesure du courant moyen
	unsigned int reference = 0;

	for (byte i = 0; i < MEASURES; ++i) {
		reference += analogRead(A0);

		if (!(i % 10))
			yield();
	}

	TRACE_DECODER("MeanCurrent %d\n", reference / MEASURES);
	yield();
	return reference / MEASURES;
}

// Write CV in Paged CV Addressing mode

// cv_adr = 1-1024
//     cv_pag = 1-255-0
//     cv_reg = 0-3
// cv_dat = 0-255
void prog_cv(unsigned int cv_adr, byte dat)
{
	byte seq_ind, a, c, cv_pag, cv_reg, cv_dat;
	unsigned int current;
	unsigned int maxcurrent = 0;
	unsigned int reference = MeanCurrent();

	cv_pag = ((cv_adr - 1) / 4) + 1;
	cv_reg = (cv_adr - 1) & 3;
	cv_dat = dat;

	for (seq_ind = 0; seq_ind < WRITE_SEQ_SIZE; seq_ind++)
	{
		switch (dcc_write_seq[seq_ind])
		{
		case DCC_PROG_FRAME_IDLE:
			dcc_tx2(22, 0xFF, 0);
			break;
		case DCC_PROG_FRAME_RESET:
			dcc_tx2(16, 0, 0);
			break;
		case DCC_PROG_FRAME_PAGE:
			dcc_tx2(22, 0b01111101, cv_pag);   //0111CRRR C=1=WR RRR=5=page
			break;
		case DCC_PROG_FRAME_DAT:
			dcc_tx2(22, 0b01111000 + cv_reg, cv_dat);  //0111CRRR C=1=WR RRR=0(cv1),1(cv2),2(cv3),3(cv4) ...

			for (int i = current = 0; i < 500; ++i) {
				if ((current = (int)((analogRead(A0) - reference) * 0.2 + current * 0.8)) > 10) {
					TRACE_DECODER("DCC_PROG_FRAME_DAT OK\n");
					seq_ind++;
					while (dcc_write_seq[seq_ind + 1] == DCC_PROG_FRAME_DAT)
						seq_ind++;

					break;
				}
			}
			break;
		case DCC_PROG_FRAME_ACK:
			dcc_tx2(16, 0, 0);

			for (int i = current = 0; i < 500; ++i) {
				if ((current = (int)((analogRead(A0) - reference) * 0.2 + current * 0.8)) > 10) {
					TRACE_DECODER("DCC_PROG_FRAME_ACK OK\n");
					return;
				}
			}
			break;
		default:
			TRACE_DECODER("DCC_PROG_FRAME_ACK no ack\n");
			break;
		}
	}
}

//--------------------------------------------------------------------------------
// Booster management
//----------------------------------------------------------------------------------

byte slot_tx = 0;

unsigned int souris_adr[MAX_WIFI_CLIENTS][MAX_CHANNELS];  // stocke les adresses des locos pour chaque souris.
										  // afin de savoir s'il y a l'adresse 1 et passer en analogique

C_slot_ch slot_ch[NB_SLOTS];

// jmd>>

void dcc_set_fct(byte slot, byte fct_num, byte val)
{
	C_slot_ch* pslot = &slot_ch[slot];

	if (pslot->fct[fct_num] != val) {
		if (fct_num <= 4) pslot->fct_0_4_modified = FCT_AUX_REPEAT;
		else if (fct_num <= 8) pslot->fct_5_8_modified = FCT_AUX_REPEAT;
		else if (fct_num <= 12) pslot->fct_9_12_modified = FCT_AUX_REPEAT;
		else if (fct_num <= 20) pslot->fct_13_20_modified = FCT_AUX_REPEAT;
		else if (fct_num <= 28) pslot->fct_21_28_modified = FCT_AUX_REPEAT;
	}

	pslot->fct[fct_num] = val;
}

void booster_maj(void)
{
	boost = BOOST_DCC;

	for (byte j = 0; j < MAX_CHANNELS; j++)
		for (byte i = 0; i < MAX_WIFI_CLIENTS; i++)
			if (souris_adr[j][i] == ADR_ANA)
				boost = BOOST_ANA;

	yield();

	if (boost != oldboost)
	{
		if (oldboost == BOOST_ANA) { analogWriteFreq(50); analogWrite(P_PIN, 0); }
		oldboost = boost;
		if (boost == BOOST_ANA)
		{
			TRACE1("BOOST ANA\n");
			analogWriteFreq(20000);
			analogWrite(P_PIN, 0);
		}

		if (boost == BOOST_DCC)
		{
			TRACE1("BOOST DCC\n");
		}
	}

	// si un mode est interdit, on peut y passer, mais on reste en AU
#if USER_DCC_OK == 0
	if (boost == BOOST_DCC) au = 1;
#endif
#if USER_ANA_OK == 0
	if (boost == BOOST_ANA) au = 1;
#endif

	if (boost == BOOST_ANA)
	{
		analogWrite(P_PIN, au ? 0 : pwm1024);
		digitalWrite(S_PIN, pwmsens ? HIGH : LOW);
	}

	if (boost == BOOST_DCC)
	{
		//maintent en cas d'arret d'urgence on coupe l'alimentation au lieu d'envoyer la trame d'arret d'urgence
		// ancienne solution
		//dcc_tx2(16, 0, 0x41); // AU : 1111111111111111 0 00000000 0 01000001 0 EEEEEEEE 1
		//notez qu'il est possible de transmettre des trames pour les acceoires par S_PIN (en dehors de cette fonction)
		if (au) {
			digitalWrite(P_PIN, LOW);
		} else {
			digitalWrite(P_PIN, HIGH);

			// GENERATION DU SIGNAL DCC pr S_PIN
			unsigned int adr;
			byte nb_a = 0;
			byte nb_b = 0;

			while (1) {
				C_slot_ch* pslot = &slot_ch[slot_tx];

				if ((adr = pslot->adr) == 0)
					nb_b++;
				else {
					nb_a++;
					byte sens = pslot->sens;
					byte vit = pslot->vit;
					byte* pfct = pslot->fct;
					byte fct04_1 = 16 * pfct[0] + 8 * pfct[4] + 4 * pfct[3] + 2 * pfct[2] + 1 * pfct[1];
					byte fct8_5 = 8 * pfct[8] + 4 * pfct[7] + 2 * pfct[6] + 1 * pfct[5];
					byte fct12_9 = 8 * pfct[12] + 4 * pfct[11] + 2 * pfct[10] + 1 * pfct[9];
					byte fct20_13 = 128 * pfct[20] + 64 * pfct[19] + 32 * pfct[18] + 16 * pfct[17] + 8 * pfct[16] + 4 * pfct[15] + 2 * pfct[14] + 1 * pfct[13];
					byte fct28_21 = 128 * pfct[28] + 64 * pfct[27] + 32 * pfct[26] + 16 * pfct[25] + 8 * pfct[24] + 4 * pfct[23] + 2 * pfct[22] + 1 * pfct[21];

					// VIT  28: 1111111111111111 0 0AAAAAAA 0 01DFSSSS 0 EEEEEEEE 1
					// VIT 126  : 1111111111111111 0 0AAAAAAA 0 00111111 0 DSSSSSSS 0 EEEEEEEE 1
#if PROTOCOLE == D17PPV3
					if (pslot->speedsteps == 28) {
						if (adr <= 127) 
							dcc_tx2(16, adr, vit28_2_dcc(sens, vit));
						else            
							dcc_tx3(16, 0xC0 + highByte(adr), lowByte(adr), vit28_2_dcc(sens, vit));
					} else {
						if (adr <= 127) 
							dcc_tx3(16, adr, 0x3f, vit126_2_dcc(sens, vit));
						else            
							dcc_tx4(16, 0xC0 + highByte(adr), lowByte(adr), 0x3f, vit126_2_dcc(sens, vit));
					}
#else
#if USER_USE_126_CRANS == 0
					if (adr <= 127) dcc_tx2(16, adr, vit28_2_dcc(sens, vit));
					else            dcc_tx3(16, 0xC0 + highByte(adr), lowByte(adr), vit28_2_dcc(sens, vit));
#else
					if (adr <= 127) dcc_tx3(16, adr, 0x3f, vit126_2_dcc(sens, vit));
					else            dcc_tx4(16, 0xC0 + highByte(adr), lowByte(adr), 0x3f, vit126_2_dcc(sens, vit));
#endif
#endif
//					pslot->vit_modified = 0;
					yield();

					// FCT 0-4: 1111111111111111 0 0AAAAAAA 0 100-FL-F4-F3-F2-F1 0 EEEEEEEE 1
					if (pslot->fct_0_4_modified) {
						if (adr <= 127)
							dcc_tx2(16, adr, 128 + fct04_1);
						else
							dcc_tx3(16, 0xC0 + highByte(adr), lowByte(adr), 128 + fct04_1);

						yield();

						if (pslot->fct_0_4_modified < 255)
							pslot->fct_0_4_modified--;
					}

					// FCT 5-8: 1111111111111111 0 0AAAAAAA 0 1011-F8-F7-F6-F5 0 EEEEEEEE 1
					if (pslot->fct_5_8_modified) {
						if (adr <= 127) 
							dcc_tx2(16, adr, 0b10110000 | fct8_5);
						else            
							dcc_tx3(16, 0xC0 + highByte(adr), lowByte(adr), 0b10110000 | fct8_5);

						yield();

						if (pslot->fct_5_8_modified < 255)
							pslot->fct_5_8_modified--;
					}

					// FCT 9-12: 1111111111111111 0 0AAAAAAA 0 1010-F12-F11-F10-F9  0 EEEEEEEE 1
					if (pslot->fct_9_12_modified) {
						if (adr <= 127) 
							dcc_tx2(16, adr, 0b10100000 | fct12_9);
						else            
							dcc_tx3(16, 0xC0 + highByte(adr), lowByte(adr), 0b10100000 | fct12_9);

						yield();

						if (pslot->fct_9_12_modified < 255)
							pslot->fct_9_12_modified--;
					}

					// FCT 13-20: 1111111111111111 0 0AAAAAAA 0 11011110 0 F20-F19-F18-F17-F16-F15-F14-F13 0 EEEEEEEE 1
					if (pslot->fct_13_20_modified) {
						if (adr <= 127) 
							dcc_tx3(16, adr, 0b11011110, fct20_13);
						else            
							dcc_tx4(16, 0xC0 + highByte(adr), lowByte(adr), 0b11011110, fct20_13);

						yield();

						if (pslot->fct_13_20_modified < 255)
							pslot->fct_13_20_modified--;
					}

					// FCT 21-28: 1111111111111111 0 0AAAAAAA 0 11011111 0 F28-F27-F26-F25-F24-F23-F22-F21 0 EEEEEEEE 1
					if (pslot->fct_21_28_modified) {
						if (adr <= 127) 
							dcc_tx3(16, adr, 0b11011111, fct28_21);
						else            
							dcc_tx4(16, 0xC0 + highByte(adr), lowByte(adr), 0b11011111, fct28_21);

						yield();

						if (pslot->fct_21_28_modified < 255)
							pslot->fct_21_28_modified--;
					}
				}

				slot_tx++;

				if (slot_tx == NB_SLOTS) 
					slot_tx = 0;
#if PRINT_PACKET
				++boosterloop;
#endif 
				if ((4 * nb_a + nb_b) >= 8) break;  // exit after 2 locos slots or 8 empty slots (or a combination)
			} // while(1)

			if (nb_a == 0)  // transmit IDLE if 0 locos updated
			{
				// IDLE   : 1111111111111111 0 11111111 0 00000000 0 EEEEEEEE 1
				dcc_tx2(16, 0xff, 0);
			}
		}
	}  //(boost==BOOST_DCC)

	delay(0);
}
