//--------------------------------------------------------------------------------
// Module D17 par Ulysse Delmas-Bgue
// USB DCC++ protocol part
//--------------------------------------------------------------------------------

#include "dccpp.h"

#if PROTOCOLE == DCCPP || PROTOCOLE == CDMRAIL
#include "dcc.h"
#include "inputs.h"
#include "slot.h"
#include "user.h"

/*
  DCC++ protocol  (uniquement les cmds, pas les cfg, list, prg ...)

	<0> -> <p0>   PWR OFF
	<1> -> <p1>   PWR ON

	<t slot adr vit sens(1=av)> -> <T slot vit sens>
	<f adr byte1 [byte2]> -> pas de return
		128+1(F1)+2(F2)+4(F3)+8(F4)+16(F0)
		176+1(F5)+2(F6)+4(F7)+8(F8)
		160+1(F9)+2(F10)+4(F11)+8(F12)
		222   byte2:1(F13)....128(F20)
		223   byte2:1(2F1)....128(F28)

	<a adr0_511 sub0_3 on0_1> -> pas de return
	<T ID 0=dir/1=dev> -> <H ID pos>
	<Z ID on> -> pas de return

	<Q> -> toutes les entrees <Q ID> low ou <q ID> high
	en temps normal seulement transmision des changements a chaque bcl
	<c> -> <a0-1023>
*/

static int acc_id;
static int acc_adr;
static byte acc_pair;

static byte serial_state = 0;
static char serial_cmd[20];
static byte serial_ind = 0;

static void dccpp_in_send_all(void)
{
	byte i, j, m;

	j = 0; 
	m = 1;

	for (i = 0; i < DCCPP_SENSOR_NB; i++)
	{
		Serial.printf("<%c %d>", (io_in[j] & m) ? 'Q' : 'q', i);
		m = m << 1;

		if (m == 0) {
			m = 1;
			j++;
		}
	}
}

void dccpp_in_send_diff(void)
{
	byte i, j, m, in;

	j = 0; 
	m = 1; 
	in = io_in[0] ^ io_in_old[0];

	for (i = 0; i < DCCPP_SENSOR_NB; i++)
	{
		if (in & m)
		{
			Serial.printf("<%c %d>", (io_in[j] & m) ? 'Q' : 'q', i);
		}

		m = m << 1;
		if (m == 0)
		{
			m = 1; j++;
			in = io_in[j] ^ io_in_old[j];
		}
	}

}

static byte cdm_nb_mod_s88_8bits = 0;
static char cdm_old_s88_reply[(IO_IN_NB / 4) + 5] = "";

// exemple de retour:
//      <y 0000>");    // LSB
//      <y 5963>");    // det0=0 puis 1 0 1   1001 0110 1100
//      <y 5963124884FFFFFFFFFFFFFF>");    // mettre l'hexa en majuscule
void cdm_in_send_diff(void)
{
	if (cdm_nb_mod_s88_8bits == 0) return;

	char reply[(IO_IN_NB / 8) + 5];
	reply[0] = '\0';

	strcat(reply, "<y ");
	byte j = 3;
	for (byte i = 0; i < cdm_nb_mod_s88_8bits; i++) {
		byte in8 = io_inmem_usb[i];
		sprintf(&reply[j], "%X", in8 & 0x0f);
		in8 = in8 >> 4;
		j++;
		sprintf(&reply[j], "%X", in8 & 0x0f);
		j++;
		io_inmem_usb[i] = 0;
	}
	strcat(reply, ">");
	if (strcmp(reply, cdm_old_s88_reply) != 0)   // send only if diff
	{
		Serial.print(reply);
		strcpy(cdm_old_s88_reply, reply);
	}
}

static byte dccpp_get_slot(unsigned int adr)
{
	byte slot = get_slot_for_adr(adr);
	if (slot == 99) {
		slot = get_free_or_oldest_slot();
		slot_ch[slot].adr = adr;
		slot_ch[slot].sens = 1; // 1=AV                       //UU: voir si on initialise ou pas
		slot_ch[slot].vit = 0;                                //UU: voir si on demande a mettre a jours la loco avec les modified=1
		for (byte j = 0; j < 29; j++) slot_ch[slot].fct[j] = 0;
	}
	set_first_slot(slot);

	return slot;
}

static void dccpp_set_sens_vit(unsigned int adr, byte sens, byte vit)
{
	byte slot = dccpp_get_slot(adr);
	slot_ch[slot].sens = sens;
	slot_ch[slot].vit = vit;
}

static void dccpp_process(void)
{
	byte i, nb, nbi, num, slot;
	unsigned int dat[4];
	unsigned char c;
	unsigned char adr;

	// extraire tous les nombres
	nb = 0; nbi = 0; num = 0;
	for (i = 0; i < serial_ind; i++) {
		c = serial_cmd[i];
		if (c >= '0' && c <= '9') { nbi = 10 * nbi + c - '0'; num = 1; } else {
			if (num) { num = 0; if (nb < 4) dat[nb] = nbi; nb++; nbi = 0; }
		}
	}
	if (num) { num = 0; if (nb < 4) dat[nb] = nbi; nb++; nbi = 0; }

	// traiter la commande

	c = serial_cmd[0];

	if (c == '0') { au = 1; Serial.print("<p0>"); }

	else if (c == '1') { au = 0; Serial.print("<p1>"); }

	else if (c == 'c') { Serial.print("<a"); Serial.print(an); Serial.print(">"); }

#if PROTOCOLE == DCCPP
	else if (c == 's') { Serial.print("<p"); Serial.print(au ^ 1); Serial.print("><iDCC++ D17 20180222><N0: SERIAL><X><X>"); }
#elif PROTOCOLE == CDMRAIL
	else if (c == 's') { Serial.print("<p"); Serial.print(au ^ 1); Serial.println("><iDCC++ BASE STATION FOR ARDUINO UNO / ARDUINO MOTOR SHIELD: V-1.2.1+S88.1 / Mar  1 2019 19:28:46><N0: SERIAL><X><X>"); }
#endif

	else if (c == 'Q') { dccpp_in_send_all(); }

	else if (c == 'S') { Serial.print("<O>"); }

	else if (c == 't' && nb == 4) {
		// dat[0] = numro du slot dccpp non utilis (le numero du slot DCC++ commence a 1, le 0 n'est que pour les actions non bouclees ...)
		// dat[1] = adr
		// dat[2] = vitesse
		// dat[3] = sens
		adr = dat[1];
		byte vit = dat[2];
		byte sens = dat[3];
		if (adr != ADR_ANA)
			if (adr >= 1 && adr <= 10239) {
				// on suppose que les commandes sont envoyees en mode 28 crans. Chnager si ce n'est pas le cas
				if (vit > USER_SPEED_STEPS) vit = USER_SPEED_STEPS;
#if USER_USE_126_CRANS == 0
				vit = map(vit, 0, USER_SPEED_STEPS, 0, 28);
#else
				vit = map(vit, 0, DCCPP_MAX_CRAN, 0, 126);
#endif

				dccpp_set_sens_vit(adr, sens, vit);   // la vitesse doit etre transmise avec le bon format ex: 0-28 en mode 28 crans ou 0-126 en mode 126 crans
			}
		Serial.print("<T "); Serial.print(dat[0]); Serial.print(" "); Serial.print(dat[2]); Serial.print(" "); Serial.print(dat[3]); Serial.print(">");
	}

	/* en mode DCC++, les fonctions ne sont transmisent qu'une seule fois ici a la reception de la commande*/
	else if (c == 'f') {
		adr = dat[0];
		if (adr != ADR_ANA)
			if (adr >= 1 && adr <= 10239) {
				if (nb == 2) // 1 byte de cmd pour F0-F4, F5-F8, F9-F12
				{
					if (adr <= 127) dcc_tx2(16, adr, dat[1]);
					else           dcc_tx3(16, 0xC0 + highByte(adr), lowByte(adr), dat[1]);
				}
				if (nb == 3) // 2 bytes de cmd pour F13-F20, F21-F28
				{
					if (adr <= 127) dcc_tx3(16, adr, dat[1], dat[2]);
					else           dcc_tx4(16, 0xC0 + highByte(adr), lowByte(adr), dat[1], dat[2]);
				}
			}
	}

	else if (c == 'T') {
#if PROTOCOLE == DCCPP
		if (nb == 3)  // configure l'accessoire ID avec adr et pair
		{
			acc_id = dat[0];
			acc_adr = dat[1];
			acc_pair = dat[2];
			Serial.print("<O>");
		}
		if (nb == 2 && acc_id == dat[0]) // commande l'accessoire ID (A chaque fois avec RocRail on a une configuration suivit de la commande)
		{
			user_dccpp_acc(acc_adr, acc_pair, dat[1]);
			Serial.print("<H "); Serial.print(dat[0]); Serial.print(" "); Serial.print(dat[1]); Serial.print(">");
		}
#endif
#if PROTOCOLE == CDMRAIL
		if (nb == 2) {
			unsigned int acc_1_2048 = dat[0];
			byte val0_1 = 0; if (dat[1]) val0_1 = 1;
			if ((acc_1_2048 >= 1) && (acc_1_2048 <= 2048)) user_cdm_acc(acc_1_2048, val0_1);
			Serial.print("<H "); Serial.print(dat[0]); Serial.print(" "); Serial.print(dat[1]); Serial.print(">");
		}
#endif
	}
	// Y est une nouvelle commande ajoutee a DCC++ pour lire les entrees du bus S88
	else if (c == 'Y')   // <Y 2 1> 2x8 1=hexa (only hex mode is supported for CDM-Rail)
	{
		if (nb == 2) {
			if (dat[1] == 1) {
				cdm_nb_mod_s88_8bits = dat[0];
				if (cdm_nb_mod_s88_8bits > (IO_IN_NB / 8)) cdm_nb_mod_s88_8bits = IO_IN_NB / 8;
				Serial.print("<o "); Serial.print(cdm_nb_mod_s88_8bits); Serial.print("*8 1>");
			}
		}
	}
}


void parse_usb_dccpp(void)
{
	int rcmd;
	unsigned char cmd;

	while (Serial.available()) {
		rcmd = Serial.read();
		if (rcmd < 0) return;
		cmd = (unsigned char)rcmd;

		if (cmd == '<') { serial_state = 1; serial_ind = 0; continue; }
		if (serial_state == 0) continue;
		if (cmd == '\n' || cmd == '\r') continue;
		if (cmd == '>') { serial_cmd[serial_ind] = '\0'; dccpp_process(); serial_state = 0; continue; }

		serial_cmd[serial_ind] = cmd;
		serial_ind++;
		if (serial_ind == 20) serial_state = 0;
	}
}

#endif
