#include "decoder.h"

#if PROTOCOLE == D17PPV2 || PROTOCOLE == D17PPV3 || PROTOCOLE == DSTATION 

#include "dcc.h"

// RESET     : 1111111111111111 0 00000000 0 00000000 0 00000000 1
// FACTORY   : 1111111111111111111111 0 01111111 0 00001000 0 01110111 1

// Service Mode Instruction Packets for Direct Mode

// CHECKCVBIT: 1111111111111111111111 0 011110AA 0 AAAAAAAA 0 11101BBB 0 EEEEEEEE 1
// WRITECVBIT: 1111111111111111111111 0 011110AA 0 AAAAAAAA 0 1111DBBB 0 EEEEEEEE 1
// VERIFCV   : 1111111111111111111111 0 011101AA 0 AAAAAAAA 0 DDDDDDDD 0 EEEEEEEE 1
// WRITECV   : 1111111111111111111111 0 011111AA 0 AAAAAAAA 0 DDDDDDDD 0 EEEEEEEE 1

constexpr auto PREAMBLE  = 22;
// first byte
constexpr auto BITMANIP  = 0b01111000;
constexpr auto VERIFYCV  = 0b01110100;
constexpr auto WRITECV   = 0b01111100;
// third byte
constexpr auto CHECKBIT0 = 0b11100000;
constexpr auto CHECKBIT1 = 0b11101000;
constexpr auto WRITEBIT  = 0b11110000;

int Decoder::m_uiResult;
bool Decoder::m_fResult = false;
char Decoder::m_cCode;

void Decoder::Reset(byte num)
{
	for (; num; num--) {
		dcc_tx2(16, 0, 0);
		delay(5);
	}
}

void Decoder::Factory(void)
{
	for (byte i = 0; i < 5; ++i) {
		dcc_tx2(22, 0b01111111, 0b00001000);
		delay(5);
	}
}

// Direct Read in Service Mode

void Decoder::DirectRead(unsigned int address)
{
	TRACE_DECODER("Decoder::DirectRead address %d\n", address);
	byte value = 0;

	m_uiResult = -1;
	// mesure du courant moyen
	unsigned int reference = MeanCurrent();
	// address range is 0-1023
	address--;
	byte a = BITMANIP + ((address >> 8) & 3);

	// lecture bit par bit
	for (byte bit = 0; bit < 8; ++bit) {
		TRACE_DECODER("Decoder::DirectRead bit %d\n", bit);
		Reset(3);

		for (byte i = 0; i < 5; ++i) {
			dcc_tx3(PREAMBLE, a, (byte)address, CHECKBIT1 + bit);
//			delay(5);

			if (ReadAck(reference, 3)) {
				value |= (1 << bit);
				break;
			}
		}
	}

	TRACE_DECODER("Decoder::DirectRead value address %d, value %d\n", address, value);
	a = VERIFYCV + ((address >> 8) & 3);
	Reset(3);
	// vrification de l'octet
	for (byte i = 0; i < 5; ++i) {
		dcc_tx3(PREAMBLE, a, (byte)address, value);

		if (ReadAck(reference, 3)) {
			m_uiResult = value;
			break;
		}
	}

	TRACE_DECODER("Decoder::DirectRead verify %d\n", m_uiResult);
}

// Direct Read Bit in Service Mode

void Decoder::DirectReadBit(unsigned int address, byte bit)
{
	TRACE_DECODER("Decoder::DirectReadBit address %d, bit %d\n", address, bit);
	// mesure du courant moyen
	unsigned int reference = MeanCurrent();
	// address range is 0-1023
	address--;
	byte a = BITMANIP + ((address >> 8) & 3);
	byte c = CHECKBIT1 + (bit & 7);
	Reset(3);
	// vrifie si bit  1
	for (byte i = 0; i < 5; ++i) {
		dcc_tx3(PREAMBLE, a, (byte)address, c);

		if ((m_uiResult = ReadAck(reference, 3)) == true)
			break;
	}
}

// Direct Verify Byte in Service Mode

void Decoder::DirectVerify(unsigned int address, byte value)
{
	TRACE_DECODER("Decoder::DirectVerify address %d, value %d\n", address, value);
	// mesure du courant moyen
	unsigned int reference = MeanCurrent();
	// address range is 0-1023
	address--;
	byte a = VERIFYCV + ((address >> 8) & 3);
	Reset(3);
	// vrification de l'octet
	for (byte i = 0; i < 5; ++i) {
		dcc_tx3(PREAMBLE, a, (byte)address, value);

		if ((m_uiResult = ReadAck(reference, 3) == true) != 0)
			break;
	}
}

// Direct Write Bit in Service Mode

void Decoder::DirectWriteBit(unsigned int address, byte bit, bool value)
{
	TRACE_DECODER("Decoder::DirectWriteBit address %d, bit %d, value %d\n", address, bit, value);
	// mesure du courant moyen
	unsigned int reference = MeanCurrent();
	// address range is 0-1023
	address--;
	byte a = BITMANIP + ((address >> 8) & 3);
	byte c = WRITEBIT + ((byte)value << 3) + (bit & 7);
	Reset(3);
	// criture d'un bit
	for (byte i = 0; i < 5; ++i) {
		dcc_tx3(PREAMBLE, a, (byte)address, c);

		if ((m_uiResult = ReadAck(reference, 6) == true) != 0)
			break;
	}
}

// Direct Write Byte in Service Mode

void Decoder::DirectWrite(unsigned int address, byte value)
{
	TRACE_DECODER("Decoder::DirectWrite address %d, value %d\n", address, value);
	// mesure du courant moyen
	unsigned int reference = MeanCurrent();
	// address range is 0-1023
	address--;
	byte a = WRITECV + ((address >> 8) & 3);
	Reset(3);
	// criture de l'octet
	for (byte i = 0; i < 5; ++i) {
		dcc_tx3(PREAMBLE, a, (byte)address, value);

		if ((m_uiResult = ReadAck(reference, 6) == true) != 0)
			break;
	}
}

bool Decoder::ReadAck(int reference, byte reset)
{
	unsigned int i;
	unsigned int current;
	unsigned int maxcurrent = 0;

	for (i = current = 0; i < 500; ++i) {
		if ((current = (unsigned int)((analogRead(A0) - reference) * 0.2 + current * 0.8)) > 30) {
			TRACE_DECODER("Decoder::ReadAck OK, current %d\n", current + reference);
			Reset(reset ? reset : 1);
			return true;
		}

		if (maxcurrent < current)
			maxcurrent = current;
	}

	TRACE_DECODER("Decoder::ReadAck fail, max current %d\n", maxcurrent + reference);
	return false;
}

unsigned int Decoder::MeanCurrent(void)
{
	constexpr auto MEASURES = 100;
	// mesure du courant moyen
	unsigned int reference = 0;

	for (byte i = 0; i < MEASURES; ++i) {
		reference += analogRead(A0);

		if (!(i % 10))
			yield();
	}

	TRACE_DECODER("Decoder::MeanCurrent %d\n", reference / MEASURES);
	yield();
	return reference / MEASURES;
}

enum {
	DCC_PROG_FRAME_IDLE,
	DCC_PROG_FRAME_RESET,
	DCC_PROG_FRAME_PAGE,
	DCC_PROG_FRAME_DAT,
	DCC_PROG_FRAME_VER,
	DCC_PROG_FRAME_ACK,
	DCC_PROG_FRAME_END
};

// Verify CV in Paged Mode Addressing mode

static byte dcc_verify_seq[] = {
	DCC_PROG_FRAME_RESET,
	DCC_PROG_FRAME_RESET,
	DCC_PROG_FRAME_RESET,

	DCC_PROG_FRAME_PAGE,
	DCC_PROG_FRAME_PAGE,
	DCC_PROG_FRAME_PAGE,
	DCC_PROG_FRAME_PAGE,
	DCC_PROG_FRAME_PAGE,

	DCC_PROG_FRAME_RESET,
	DCC_PROG_FRAME_RESET,
	DCC_PROG_FRAME_RESET,
	DCC_PROG_FRAME_RESET,
	DCC_PROG_FRAME_RESET,
	DCC_PROG_FRAME_RESET,
	DCC_PROG_FRAME_RESET,
	DCC_PROG_FRAME_RESET,
	DCC_PROG_FRAME_RESET,

	DCC_PROG_FRAME_VER,
	DCC_PROG_FRAME_VER,
	DCC_PROG_FRAME_VER,
	DCC_PROG_FRAME_VER,
	DCC_PROG_FRAME_VER,

	DCC_PROG_FRAME_ACK,
	DCC_PROG_FRAME_ACK,
	DCC_PROG_FRAME_ACK,
	DCC_PROG_FRAME_ACK,
	DCC_PROG_FRAME_ACK,
	DCC_PROG_FRAME_ACK,

	DCC_PROG_FRAME_END
};

#define VERIFY_SEQ_SIZE sizeof(dcc_verify_seq) / sizeof(dcc_verify_seq[0])

bool Decoder::PagedVerify(byte cv_pag, byte cv_reg, byte value, unsigned int reference)
{
	byte seq_ind, a, c;

	for (seq_ind = 0; seq_ind < VERIFY_SEQ_SIZE; seq_ind++) {
		switch (dcc_verify_seq[seq_ind]) {
		case DCC_PROG_FRAME_IDLE:
			dcc_tx2(PREAMBLE, 0xFF, 0);
			break;
		case DCC_PROG_FRAME_RESET:
			Reset(1);
			break;
		case DCC_PROG_FRAME_PAGE:
			dcc_tx2(PREAMBLE, 0b01111101, cv_pag);   //0111CRRR C=1=WR RRR=5=page
			break;
		case DCC_PROG_FRAME_VER:
			dcc_tx2(PREAMBLE, 0b01110000 + cv_reg, value);  //0111CRRR C=0=WR RRR=0(cv1),1(cv2),2(cv3),3(cv4) ...

			if (ReadAck(reference, 0)) {
				TRACE_DECODER("DCC_PROG_FRAME_VER OK\n");
				return true;
			}
			break;
		case DCC_PROG_FRAME_ACK:
			if (ReadAck(reference, 1)) {
				TRACE_DECODER("DCC_PROG_FRAME_ACK OK\n");
				return true;
			}
			break;
		default:
			TRACE_DECODER("DCC_PROG_FRAME_ACK no ack\n");
			break;
		}
	}

	return false;
}

void Decoder::PagedRead(unsigned int address)
{
	byte seq_ind, a, c, cv_pag, cv_reg;
	unsigned int value;
	unsigned int reference = MeanCurrent();

	TRACE_DECODER("Decoder::PagedRead address %d\n", address);

	cv_pag = ((address - 1) / 4) + 1;
	cv_reg = (address - 1) & 3;

	for (value = 0; value < 256; ++value) {
		TRACE_DECODER("Decoder::PagedRead %d, value %d\n", address, value);
		if (PagedVerify(cv_pag, cv_reg, value, reference)) {
			m_uiResult = value;
			return;
		}

		yield();
	}

	m_uiResult = -1;
}

void Decoder::PagedVerify(unsigned int address, byte value)
{
	byte seq_ind, a, c, cv_pag, cv_reg;
	unsigned int reference = MeanCurrent();

	TRACE_DECODER("Decoder::PagedVerify %d\n", value);
	cv_pag = ((cv_adr - 1) / 4) + 1;
	cv_reg = (cv_adr - 1) & 3;
	m_uiResult = (unsigned int) PagedVerify(cv_pag, cv_reg, value, reference);
}

static byte dcc_write_seq[] = {
	DCC_PROG_FRAME_RESET,
	DCC_PROG_FRAME_RESET,
	DCC_PROG_FRAME_RESET,

	DCC_PROG_FRAME_PAGE,
	DCC_PROG_FRAME_PAGE,
	DCC_PROG_FRAME_PAGE,
	DCC_PROG_FRAME_PAGE,
	DCC_PROG_FRAME_PAGE,

	DCC_PROG_FRAME_RESET,
	DCC_PROG_FRAME_RESET,
	DCC_PROG_FRAME_RESET,
	DCC_PROG_FRAME_RESET,
	DCC_PROG_FRAME_RESET,
	DCC_PROG_FRAME_RESET,
	DCC_PROG_FRAME_RESET,
	DCC_PROG_FRAME_RESET,
	DCC_PROG_FRAME_RESET,

	DCC_PROG_FRAME_DAT,
	DCC_PROG_FRAME_DAT,
	DCC_PROG_FRAME_DAT,
	DCC_PROG_FRAME_DAT,
	DCC_PROG_FRAME_DAT,

	DCC_PROG_FRAME_ACK,
	DCC_PROG_FRAME_ACK,
	DCC_PROG_FRAME_ACK,
	DCC_PROG_FRAME_ACK,
	DCC_PROG_FRAME_ACK,
	DCC_PROG_FRAME_ACK,

	DCC_PROG_FRAME_END
};

#define WRITE_SEQ_SIZE sizeof(dcc_write_seq) / sizeof(dcc_write_seq[0])

void Decoder::PagedWrite(unsigned int address, byte value)
{
	byte seq_ind, a, c, cv_pag, cv_reg;
	unsigned int current;
	unsigned int maxcurrent = 0;
	unsigned int reference = MeanCurrent();

	TRACE_DECODER("Decoder::PagedWrite %d, value %d\n", address, value);
	cv_pag = ((cv_adr - 1) / 4) + 1;
	cv_reg = (cv_adr - 1) & 3;

	for (seq_ind = 0; seq_ind < WRITE_SEQ_SIZE; seq_ind++) {
		switch (dcc_write_seq[seq_ind]) {
		case DCC_PROG_FRAME_IDLE:
			dcc_tx2(PREAMBLE, 0xFF, 0);
			break;
		case DCC_PROG_FRAME_RESET:
			Reset(1);
			break;
		case DCC_PROG_FRAME_PAGE:
			dcc_tx2(PREAMBLE, 0b01111101, cv_pag);   //0111CRRR C=1=WR RRR=5=page
			break;
		case DCC_PROG_FRAME_DAT:
			dcc_tx2(PREAMBLE, 0b01111000 + cv_reg, value);  //0111CRRR C=1=WR RRR=0(cv1),1(cv2),2(cv3),3(cv4) ...

			if (ReadAck(reference, 0)) {
				TRACE_DECODER("DCC_PROG_FRAME_DAT OK\n");
				m_uiResult = 1;
				return;
			}
			break;
		case DCC_PROG_FRAME_ACK:
			if (ReadAck(reference, 1)) {
				TRACE_DECODER("DCC_PROG_FRAME_ACK OK\n");
				m_uiResult = 1;
				return;
			}
			break;
		default:
			TRACE_DECODER("DCC_PROG_FRAME_ACK no ack\n");
			break;
		}
	}

	m_uiResult = 0;
}
#endif