#include "config.h"

#if PROTOCOLE == DSTATION

#include "decoder.h"
#include "inputs.h"
#include "leds.h"
#include "slot.h"
#include "user.h"

static byte ds_get_slot(unsigned int adr)
{
	byte slot = get_slot_for_adr(adr);
	if (slot == 99) {
		slot = get_free_or_oldest_slot();
		slot_ch[slot].adr = adr;
		slot_ch[slot].sens = 1; // 1=AV                       //UU: voir si on initialise ou pas
		slot_ch[slot].vit = 0;                                //UU: voir si on demande a mettre a jours la loco avec les modified=1
		for (byte j = 0; j < 29; j++) slot_ch[slot].fct[j] = 0;
	}
	set_first_slot(slot);

	return slot;
}

static void ds_set_sens(unsigned int adr, byte sens)
{
	byte slot = ds_get_slot(adr);
	slot_ch[slot].sens = sens;
}

static void ds_set_vit(unsigned int adr, byte vit)
{
	byte slot = ds_get_slot(adr);
	slot_ch[slot].vit = vit;
}
// FCT 0-4  : 1111111111 0 0AAAAAAA 0 100-FL-F4-F3-F2-F1  0 EEEEEEEE 1
// FCT 5-8  : 1111111111 0 0AAAAAAA 0 1011-F8-F7-F6-F5    0 EEEEEEEE 1
// FCT 9-12 : 1111111111 0 0AAAAAAA 0 1010-F12-F11-F10-F9 0 EEEEEEEE 1
// FCT 13-20: 1111111111 0 0AAAAAAA 0 11011110            0 F20-F19-F18-F17-F16-F15-F14-F13 0 EEEEEEEE 1
// FCT 11-28: 1111111111 0 0AAAAAAA 0 11011111            0 F28-F27-F26-F25-F24-F23-F22-F21 0 EEEEEEEE 1
static void ds_set_fct(unsigned int adr, byte fct, byte val)
{
	byte slot = ds_get_slot(adr);
	dcc_set_fct(slot, fct, val);
}
/*
	if(val) dcc_ch2[i].fct |=  (1 << fct);    //UU: si un jour, on passe en unsigned long
	else    dcc_ch2[i].fct &= ~(1 << fct);

	unsigned long fctl = dcc_ch2[i].fct;
	if(fct>=0 && fct<=4)
	{
		byte tx = 0x80 | ((byte)(fctl>>1)&0x0f);
		if(fctl & (1<<0)) tx |= 16;
		dcc_tx2(16, adr, tx);
		yield();
	}
	if(fct>=5 && fct<=8)
	{
		byte tx = 0xB0 | ((byte)(fctl>>5)&0x0f);
		dcc_tx2(16, adr, tx);
		yield();
	}
	if(fct>=9 && fct<=12)
	{
		byte tx = 0xA0 | ((byte)(fctl>>9)&0x0f);
		dcc_tx2(16, adr, tx);
		yield();
	}
	if(fct>=13 && fct<=20)
	{
		byte tx = ((byte)(fctl>>13)&0xff);
		dcc_tx3(16, adr, 0xDE, tx);
		yield();
	}
	if(fct>=21 && fct<=28)
	{
		byte tx = ((byte)(fctl>>21)&0xff);
		dcc_tx3(16, adr, 0xDF, tx);
		yield();
	}
	*/

	// --> 100 Ready\n
	// setPing()\n --> @200 Ok\n
	// setPower(1)\n --> 200 Ok\n
	// setLocoSpeed(49174,0,0) --> 200 Ok\n               //0xC000 + DCC ADR, 0-1023, 0=DCC28 2=DCC126
	// setLocoDirection(49174,1) --> 200 Ok\n             //1=AV 2=AR
	// setLocoFunction(49174,9,1) --> 200 Ok\n            //F9=1  F0-F28 supported
	// setTurnout(14336,1) --> 200 Ok\n                   //Pair nb 1=0x3800 2044=0x3ffc
	// getS88(2)\n --> @S88,0-65535,0-65535\n200 Ok\n     //max 32 modules de 16
	// setLocoConfig(0-49152,0-1023,0-255)\n --> 200 Ok\n //adr

static char ds_serial_cmd[100];
static byte ds_serial_ind = 0;

static byte ds_arg_nb = 0;
static unsigned int ds_arg[5] = { 0, 0, 0, 0, 0 };

static void desktop_station_decompose(void)
{
	byte i = 0;
	ds_arg_nb = 0;
	unsigned int v = 0;

	while (char c = ds_serial_cmd[i++]) {
		if (c >= '0' && c <= '9') v = 10 * v + c - '0';
		if (c == ',' || c == ')') if (ds_arg_nb <= 4) { ds_arg[ds_arg_nb] = v; ds_arg_nb++; }
		if (c == '(' || c == ',') v = 0;
	}
}

static byte firstPing = 1;

static void desktop_station_process(void)
{
	desktop_station_decompose();
	TRACER.printf("%s ", ds_serial_cmd);

	if (!strncmp(ds_serial_cmd, "setPower", 8) && ds_arg_nb == 1) {
		TRACE1("%d\n", ds_arg[0]);
		if (ds_arg[0] == 1) au = 0; else au = 1;
	} else if (!strncmp(ds_serial_cmd, "setLocoDirection", 16) && ds_arg_nb == 2) {
		unsigned int adr = ds_arg[0] - 0xC000;
		TRACE1("%d %d\n", adr, ds_arg[1]);
		if (adr != ADR_ANA)
			if (adr >= 1 && adr <= 10239)
			{
				if (ds_arg[1] == 1) ds_set_sens(adr, 0); // AV
				if (ds_arg[1] == 2) ds_set_sens(adr, 1); // AR
			}
	} else if (!strncmp(ds_serial_cmd, "setLocoSpeed", 12) && ds_arg_nb == 3) {
		unsigned int adr = ds_arg[0] - 0xC000;
		TRACE1("%d %d\n", adr, ds_arg[1]);
		if (adr != ADR_ANA)
			if (adr >= 1 && adr <= 10239) {
#if USER_USE_126_CRANS == 0
				ds_set_vit(adr, map(ds_arg[1], 0, 1023, 0, 28));
#else
				ds_set_vit(adr, map(ds_arg[1], 0, 1023, 0, 126));
#endif
			}
	} else if (!strncmp(ds_serial_cmd, "setLocoFunction", 15) && ds_arg_nb == 3) {
		unsigned int adr = ds_arg[0] - 0xC000;
		TRACE1("%d %d %d\n", adr, ds_arg[1], ds_arg[2]);
		if (adr != ADR_ANA)
			if (adr >= 1 && adr <= 10239) {
				ds_set_fct(adr, ds_arg[1], ds_arg[2]);   // adr, fct_0_28, val_0_1

			}
	} else if (!strncmp(ds_serial_cmd, "setLocoConfig", 13) && ds_arg_nb == 3) {
		TRACE1("%d %d %d\n", ds_arg[0], ds_arg[1], ds_arg[2]);
		// sequence de programmation dcc
		led_sign_of_life(1); //LED ON
		// programing seqeunce
		prog_cv(ds_arg[1], ds_arg[2]);  // cv_adr, cv_dat
		// disable booster during 1s to reboot decoder
		digitalWrite(P_PIN, LOW);
		digitalWrite(S_PIN, LOW);
		led_sign_of_life(0);
		delay(1000); //UU: possible to check with can
		digitalWrite(P_PIN, HIGH);
	} else if (!strncmp(ds_serial_cmd, "getLocoConfig", 13) && ds_arg_nb == 2) {
		TRACE1("%d %d\n", ds_arg[0], ds_arg[1]);
		// sequence de programmation dcc
		led_sign_of_life(1); //LED ON
		Decoder::DirectRead(ds_arg[1]);	// cv_adr
		// disable booster during 1s to reboot decoder
		digitalWrite(P_PIN, LOW);
		digitalWrite(S_PIN, LOW);
		led_sign_of_life(0);
		delay(1000); //UU: possible to check with can
		digitalWrite(P_PIN, HIGH);
		TRACE1("CV %d value = %d\n", ds_arg[1], Decoder::m_uiResult);
		Serial.printf("%d\n", Decoder::m_uiResult);
	} else if (!strncmp(ds_serial_cmd, "setTurnout", 10) && ds_arg_nb == 2) {
		unsigned int acc_1_320 = ds_arg[0] - 0x3800;
		byte val0_1 = 0; 
		if (ds_arg[1]) val0_1 = 1;
		TRACE1("%d %d\n", acc_1_320, val0_1);

		if ((acc_1_320 >= 1) && (acc_1_320 <= 320)) 
			user_ds_acc(acc_1_320, val0_1);
	} else if (!strncmp(ds_serial_cmd, "getS88", 6) && ds_arg_nb == 1) {
		byte nb = ds_arg[0];
		TRACE1("%d\n", nb);
		Serial.print("@S88,");
		for (byte i = 0; i < nb; i++) {
			unsigned int in16;
			in16 = 0;
			if (i < (IO_IN_NB / 16))   //6x16=96 entrees max
			{
				//in16 = (io_in[2*i+1] << 8) | io_in[2*i];          // sans memorisation
				in16 = (io_inmem_usb[2 * i + 1] << 8) | io_inmem_usb[2 * i];  // avec memorisation pour
				io_inmem_usb[2 * i + 1] = io_inmem_usb[2 * i] = 0;            // ne pas louper les impulsions
			}
			Serial.print(in16, HEX); Serial.print(",");
		}
		Serial.println("");
	} else if (!strncmp(ds_serial_cmd, "setPing", 7)) {
		TRACE1("\n");
		if (firstPing) { firstPing--; Serial.print("300 Command error\n"); return; }
		Serial.println("@DSG,01,00,00,02,00,00,00,00");
	} else {
		TRACE1("Unknown command %s\n", ds_serial_cmd);
		Serial.print("300 Command error\n");
		return;
	}

	Serial.print("200 Ok\n");
}

void parse_usb_desktop_station(void)
{
	int rcmd;
	unsigned char cmd;

	while (Serial.available()) {
		rcmd = Serial.read();
		if (rcmd < 0) return;
		cmd = (unsigned char)rcmd;
		if (cmd == '\r') continue;
		if (cmd == '\t') continue;
		if (cmd == ' ') continue;

		if (cmd == '\n') { 
			ds_serial_cmd[ds_serial_ind] = '\0'; 
			desktop_station_process(); 
			ds_serial_ind = 0; 
			continue; 
		} else { 
			ds_serial_cmd[ds_serial_ind] = cmd; 
			
			if (ds_serial_ind < 98) 
				ds_serial_ind++; 
		}
	}
}

#endif
