//--------------------------------------------------------------------------------
// Module D17 par Ulysse Delmas-Bgue
// Leds (128) via MAX7219 IC on SPI bus
//----------------------------------------------------------------------------------

#include "leds.h"

#if USER_USE_MAX

byte io_led[IO_LED_NB / 8] = { 0, 0, 0, 0 ,0, 0, 0, 0,    0, 0, 0, 0 ,0, 0, 0, 0 };  // possibilite d'initialiser les valeurs
byte io_cli[IO_LED_NB / 8] = { 0, 0, 0, 0 ,0, 0, 0, 0,    0, 0, 0, 0 ,0, 0, 0, 0 };  // ex led clignotante
byte io_pha[IO_LED_NB / 8] = { 0, 0, 0, 0 ,0, 0, 0, 0,    0, 0, 0, 0 ,0, 0, 0, 0 };
byte led_cpt = 0;


void user_led(byte num, byte val)
{
    byte msk;
    msk = 1 << (num & 7);
    if(val) io_led[num / 8] |= msk;
    else    io_led[num / 8] &= (0xff ^ msk);    
}


void user_led_cli(byte num, byte val)
{
    byte msk;
    msk = 1 << (num & 7);
    if(val) io_cli[num / 8] |= msk;
    else    io_cli[num / 8] &= (0xff ^ msk);
}


void user_led_pha(byte num, byte val)
{
    byte msk;
    msk = 1 << (num & 7);
    if(val) io_pha[num / 8] |= msk;
    else    io_pha[num / 8] &= (0xff ^ msk);
}


void led_shift8(byte dat)
{
    byte i;

    pinMode(DAT_PIN, OUTPUT);

    for(i = 0; i < 8; i++)
    {
         if(dat & 0x80) digitalWrite(DAT_PIN, HIGH);
         else           digitalWrite(DAT_PIN, LOW);
             
         delayMicroseconds(4);
         digitalWrite(CLK_PIN, HIGH);
         delayMicroseconds(4);  //periode=8us = 125KHz (ok sur longue distances)
         digitalWrite(CLK_PIN, LOW);
         
         dat = dat<<1;
    }
}

  
static void led_reg(byte reg0, byte dat0, byte reg1, byte dat1)
{
     digitalWrite(CLK_PIN, LOW);
     digitalWrite(LLED_PIN, HIGH); delayMicroseconds(10); digitalWrite(LLED_PIN, LOW); delayMicroseconds(10);
     led_shift8(reg1);
     led_shift8(dat1);
     led_shift8(reg0);
     led_shift8(dat0);
     digitalWrite(LLED_PIN, HIGH); delayMicroseconds(10); 
     yield();
}


static byte compute_led_on_cli_phase(byte on, byte cli, byte phase)
{
    byte led;

    if(led_cpt & 1) led = 0xff;  else led = 0;   //clignotement 1Hz
    led ^= phase;                                //changement de phase
    led |= ~cli;                                 //allumer si pas clignotant
    led &= on;                                   //eteindre si pas on

    return led;
}


void led_maj(void)
{
    byte i, dat0, dat1;
    
    for(i = 0; i < 8; i++)
    {
        dat0 = compute_led_on_cli_phase(io_led[    i], io_cli[    i], io_pha[    i]);
        dat1 = compute_led_on_cli_phase(io_led[8 + i], io_cli[8 + i], io_pha[8 + i]);
        led_reg(1 + i, dat0, 1 + i, dat1);
    }
}


void led_init(void)
{
                              // R0    : bypass
                              // R1-8  : data
    led_reg(9  ,  0,  9,  0); // R9    : decode mode off (1ere fois pas sure que ca marche) 
    led_reg(9  ,  0,  9,  0); // R9    : decode mode off
    led_reg(0xa,0x4,0xa,0x4); // R10   : intensite 4/15 (0xf=max)
    led_reg(0xb,  7,0xb,  7); // R11   : 8 digit    
    led_reg(0xc,  1,0xc,  1); // R12   :shutdown mode off
                              // R13-14: NA
    led_reg(0xf,  0,0xf,  0); // R15   :test mode off (1=on)

    led_maj(); 
}
#endif

void led_sign_of_life(byte val)
{
#if USER_USE_MAX == 0
    digitalWrite(LED_PIN, val ? LOW : HIGH);  //active LOW
#elif USER_CLI_LED_MAX != -1
    user_led(USER_CLI_LED_MAX, val);
#endif
}

