//------------------------------------------------------------------------------
// Module D17 par Ulysse Delmas-Bgue et Jean-Michel Dubois pour 3 lignes
// PCA9685 (3x16=48)
//------------------------------------------------------------------------------

#include "pca9685.h"
#include "i2c.h"

#if USER_USE_PCA

#include "i2c.h"

constexpr auto PCA9685_MODE1_REG = 0x00;
constexpr auto PCA9685_MODE2_REG = 0x01;
constexpr auto PCA9685_LED0_REG = 0x06;
constexpr auto PCA9685_PRESCALE_REG = 0xfe;
constexpr auto PCA9685_BASE = 0x80;
constexpr auto UPDATE = 1;
constexpr auto SERVO = 2;

static byte pca9685_tx_srv[ IO_PCA9685_OUT_NB];  // bit0=tx  bit1=srv
static unsigned int pca9685_on[ IO_PCA9685_OUT_NB];
static unsigned int pca9685_off[IO_PCA9685_OUT_NB];
static unsigned int pca9685_off_target[IO_PCA9685_OUT_NB];
static byte srv_speed[IO_PCA9685_OUT_NB];
static bool i2c_pca9685_online[IO_PCA9685_NB];


static bool set_pca9685_reg(byte i2c_adr, byte adr, byte dat)
{
    i2c_start();
    if (! i2c_wr_8(PCA9685_BASE + 2 * i2c_adr + I2C_WRITE))  //A6=1,A5-A0=pin,R/W#
	{
        return false;
    }
    if (! i2c_wr_8(adr))
    {
        return false;
    }
    if (! i2c_wr_8(dat))
    {
        return false;
    }
    i2c_stop();  
    return true;
}


static void set_pca9685_on_off(byte i2c_adr, byte out, unsigned int on, unsigned int off)
{
    byte on_h, on_l, off_h, off_l;
    
    on_h  = byte(on  >> 8); on_l  = byte(on  & 255);  //existe aussi lowByte highByte
    off_h = byte(off >> 8); off_l = byte(off & 255); 

    i2c_start();  // A6=1,A5-A0=pin,R/W#
    i2c_wr_8(PCA9685_BASE + 2 * i2c_adr + I2C_WRITE);
    i2c_wr_8(PCA9685_LED0_REG + 4 * out);
    i2c_wr_8(on_l);
    i2c_wr_8(on_h);
    i2c_wr_8(off_l);
    i2c_wr_8(off_h);
    i2c_stop();
    yield();
}

/*
static void set_pca9685_only_off_out(byte i2c_adr, byte out, unsigned int off)
{
    byte off_h, off_l;
    
    off_h = byte(off >> 8); off_l = byte(off & 255); 

    i2c_start();  // A6=1,A5-A0=pin,R/W#
    i2c_wr_8(PCA9685_BASE + 2 * i2c_adr + I2C_WRITE);
    i2c_wr_8(PCA9685_LED0_REG + 4 * out + 2);
    i2c_wr_8(off_l);
    i2c_wr_8(off_h);
    i2c_stop();
    yield();
}
*/

// only used at init
static void set_pca9685_pwm0_100(byte num, byte pwm)
{
    byte i2c_adr;
    byte out;
    unsigned int off;
    
    i2c_adr = num >> 4; 
    out = num & 15;
    off = map(pwm, 0, 100, 0, 4096); //c bien 4096 et pas 4095 !
    
    if(off == 0)         set_pca9685_on_off(i2c_adr, out,    0, 4096);
    else if(off == 4096) set_pca9685_on_off(i2c_adr, out, 4096,    0);
    else                 set_pca9685_on_off(i2c_adr, out,    0, off );
}


/* no more used
static void set_paca9685_srv50_250(byte num, byte pulse)  //50 -> 150 -> 250
{
    byte i2c_adr;
    byte out;
    unsigned int off;
    
    i2c_adr = byte(num >> 4); 
    out = byte(num & 15);
    off = map(pulse, 50, 250, 102, 512);
    
    set_pca9685_on_off(i2c_adr, out, 0, off );
}
*/

void pca9685_init(void)
{
    byte i;  
    byte i2c_adr;

    for(i2c_adr = 0; i2c_adr < IO_PCA9685_NB; i2c_adr++)
    {
        // reset et config MODE1
        i2c_pca9685_online[i2c_adr] = set_pca9685_reg(i2c_adr, PCA9685_MODE1_REG, 0xB0); //reset=1, extclk=0, autoincrement=1, sleep=1 (all leds are off after reset)
        if(! i2c_pca9685_online[i2c_adr]) continue;
		TRACE1(" pca9685 %d found\n", i2c_adr);
        delay(1);

        // Frequence de decoupage
        // 50Hz pour les servos: 25.000.000/(4096*50Hz) - 1 = 122-1 = 121.
        // 60Hz pour les servos: 25.000.000/(4096*60Hz) - 1 = 102-1 = 101.
        // mesure:
        // 121 -> 18.7ms     -1.3ms     err -6.5%
        // 130 -> 19.92ms    -0.08ms    err  0.4%
        // 131 -> 20.07ms    +0.07ms    err  0.4%
        set_pca9685_reg(i2c_adr, PCA9685_PRESCALE_REG, 130);

        // exit from sleep
        set_pca9685_reg(i2c_adr, PCA9685_MODE1_REG, 0x20);    //reset=0, extclk=0, autoincrement=1, sleep=0
        delay(1); //wait exit from low power mode

        // config MODE2
        set_pca9685_reg(i2c_adr, PCA9685_MODE2_REG, 0x04);     // invert=0 change=onSTOP out=totem off=led_0
    }
        
    /* toutes les sorties a 0 */
    for(i = 0; i < IO_PCA9685_OUT_NB; i++) 
    {
        if(! i2c_pca9685_online[i / 16]) continue;
        
        set_pca9685_pwm0_100(i, 0);      //uu: voir si on fait pas avant d'activer les sorties
        pca9685_on[i] = 0;
        pca9685_off[i] = 4096;
        pca9685_tx_srv[i] = 0; 
        srv_speed[i] = 0;    
    }

    /* spares  
    set_pca9685_on_off(0,0 ,4096,0); //out 0 always on
    set_pca9685_on_off(0,1 ,0,4096); //out 0 always off
    set_pca9685_on_off(0,8 ,0,2048); //out 50%
    set_pca9685_on_off(0,9 ,0,100);  //servo 0.5ms   102  -> ok LA 457us
    set_pca9685_on_off(0,10,0,200);  //servo 1.0ms   206
    set_pca9685_on_off(0,11,0,300);  //servo 1.5ms   308  -> ok LA 1.46ms
    set_pca9685_on_off(0,12,0,400);  //servo 2.0ms   410  
    set_pca9685_on_off(0,13,0,500);  //servo 2.5ms   512*4.882us=2.5ms    
    set_paca9685_pwm0_100(0, 0);
    set_paca9685_srv50_250(11, 150);
    */
}


//variation lente des servos
void pca9685_slow_srv(void)
{
    byte num;
    unsigned int off, off_target;

    for(num = 0; num < IO_PCA9685_OUT_NB; num++)
    {
        if(! i2c_pca9685_online[num / 16]) continue; // PCA9685 module is not present 

        if((pca9685_tx_srv[num] & 2) == 0) continue; // servos only
        
        if(pca9685_off_target[num] == pca9685_off[num]) continue;
        
        off_target = pca9685_off_target[num];
        off = pca9685_off[num];
        
        if(srv_speed[num]==0) 
        { 
            off = off_target;
        }
        else
        {
            if(off_target > off) { off += srv_speed[num]; if(off > off_target) off = off_target; }
            if(off_target < off) { off -= srv_speed[num]; if(off < off_target) off = off_target; }
        }
        if(pca9685_off[num] != off)
        {
            pca9685_off[num] = off;
            pca9685_tx_srv[num] |= UPDATE;  //update
        }        
    }   
}


// return 1 when update is finished
byte pca9685_maj(byte index)
{
    byte i, num;
    byte i2c_adr;
    byte out;

    for(i = 0; i < 4; i++)
    {
        num = 4 * index + i;
        if(num >= IO_PCA9685_OUT_NB) return 1;       
        
        if(! i2c_pca9685_online[num / 16]) continue;   // PCA9685 module is not present

        if(pca9685_tx_srv[num] & UPDATE)
        {
            pca9685_tx_srv[num] &= ~UPDATE;

            i2c_adr = num / 16; 
            out = num & 15;

            set_pca9685_on_off(i2c_adr, out, pca9685_on[num], pca9685_off[num]);
        }
    }

    return 0;
}


void user_pwm_0_100(byte num, byte val)
{
    unsigned int off;
    
    if(num >= IO_PCA9685_OUT_NB) return;

    off = map(val, 0, 100, 0, 4096);

    if(off == 0)         { pca9685_on[num] = 0;    pca9685_off[num] = 4096; }
    else if(off == 4096) { pca9685_on[num] = 4096; pca9685_off[num] =    0; }
    else                 { pca9685_on[num] = 0;    pca9685_off[num] =  off; }

    pca9685_tx_srv[num] |= UPDATE;
    pca9685_tx_srv[num] &= ~SERVO;  //not a servo
}


void user_servo_500_2500(byte num, unsigned int val)
{   
    unsigned int off;

    if(num >= IO_PCA9685_OUT_NB) return;

    off = map(val, 500, 2500, 102, 512);    //4096=20ms > 102=0.5ms 512=2.5ms

    pca9685_on[num] = 0;    pca9685_off_target[num] = off;

    pca9685_tx_srv[num] |= SERVO;  //is a servo (no update, done in pca9685_slow_srv)
  
}


void user_servo_speed(byte num, byte speed)
{
    if(num >= IO_PCA9685_OUT_NB) return;

    srv_speed[num] = speed; 
}

#endif

