//----------------------------------------------------------------------------------
// Module D17 par Ulysse Delmas-Bègue et à la marge Jean-Michel Dubois
// Mouse protocol part
//----------------------------------------------------------------------------------

#include "config.h"

#if PROTOCOLE <= D17PPV3

#include "protocol.h"
#include "dcc.h"
#include "inputs.h"
#include "leds.h"
#include "pca9685.h"
#include "signal.h"
#include "slot.h"
#include "turnout.h"
#include "user.h"
#include "uservars.h"

#if PROTOCOLE == D17PPV2 || PROTOCOLE == D17PPV3
#include "speed.h"
#include "decoder.h"
#endif

#if PROTOCOLE == D17PPV3
#include "block.h"

bool fUseBlock = false;					// true si au moins un client supporte le suivi des trains
byte uiClientProtocol[MAX_WIFI_CLIENTS];		// niveau de protocole utilisé par le client
#endif

char status_tx[256];
char tx[256];
#if USE_TRACER == 1
char status_tx_old[256];
#endif
byte status_txed[MAX_WIFI_CLIENTS];

static byte hexdigit(byte v)
{
    return (v < 10) ? '0' + v : 'A' + v - 10;
}

static void update_in_status_message(byte ch)
{
#if USER_USE_S88 || USER_USE_MCPIN
    byte k;
    char* ptx;

    for (k = 0, ptx = status_tx + 9; k < IO_IN_NB / 8; k++) {
        *ptx++ = hexdigit(io_inmem_wifi[ch][k] & 0x0f);
        *ptx++ = hexdigit(io_inmem_wifi[ch][k] >> 4);
        io_inmem_wifi[ch][k] = 0; // clr
    }
#else
	strcpy(status_tx + 9, "000000000000000000000000");
#endif
}

static void build_status_message(byte ch)
{
    byte k;
    char* ptx = status_tx;

    // au[0/1] ana/dcc iXXXXXXXXXXXXXXXXXXXXXXXX dXXXXXXXXXXXX eXXXXXXXXXXXX uXXXXXXXXXXXXXXXXXXXXXXXX cYY gXXXXXXX[XXXXXXX[..]]
    // 01 2   3456    789012345678901234567890123456789012345678901234567890123456789012345678901234567890123456789012
    // 0                 1         2         3         4         5         6         7         8         9          10

    // emergency stop
    strcpy(ptx, au ? "au1 " : "au0 ");
    ptx += 4;
    // dcc/analog mode
    strcpy(ptx, boost == BOOST_ANA ? "ana i" : "dcc i");
    ptx += 5;
#if USER_USE_S88 || USER_USE_MCPIN
    // inputs
    for (k = 0; k < IO_IN_NB / 8; k++) {
        *ptx++ = hexdigit(io_in[k] & 0x0f);
        *ptx++ = hexdigit(io_in[k] >> 4);
    }

    *ptx++ = ' ';
#else
	strcpy(ptx, "000000000000000000000000 ");
	ptx += 25;
#endif
    // direct turnouts
    *ptx++ = 'd';

    for (k = 0; k < IO_AIG_NB; k += 4) {
        *ptx++ = hexdigit(8 * io_aig_dir[k + 3] + 4 * io_aig_dir[k + 2] + 2 * io_aig_dir[k + 1] + io_aig_dir[k]);
    }

    *ptx++ = ' ';
    // deviated turnouts
    *ptx++ = 'e';

    for (k = 0; k < IO_AIG_NB; k += 4) {
        *ptx++ = hexdigit(8 * io_aig_dev[k + 3] + 4 * io_aig_dev[k + 2] + 2 * io_aig_dev[k + 1] + io_aig_dev[k]);
    }

    *ptx++ = ' ';
    // user variables
    *ptx++ = 'u';

    for (k = 0; k < VAR_U_NB / 8; k++) {
        *ptx++ = hexdigit(var_u[k] & 0x0f);
        *ptx++ = hexdigit(var_u[k] >> 4);
    }

    // current measure
    sprintf(ptx, " c%02d", (int)map(an, 0, 1023, 0, 99));
    ptx += 4;
#if PROTOCOLE == D17PPV2 || PROTOCOLE == D17PPV3
	// speed measure
	if (Speed::m_fMesureFaite) {
		Speed::m_fMesureFaite = false;
		ptx += sprintf(ptx, " M%02X%02X%08x", Speed::m_ucTrain, Speed::m_ucPas, Speed::m_uiTemps);
		au = 1;
	}
	// CV read, write, verify
	if (Decoder::m_fResult) {
		Decoder::m_fResult = false;
		ptx += sprintf(ptx, " P%c%04X", Decoder::m_cCode, Decoder::m_uiResult);
	}
#if PROTOCOLE == D17PPV3
	if (uiClientProtocol[ch] > 2) {
		// changed signals
		Signal* pSignal;
		*ptx++ = ' ';
		*ptx++ = 'S';

		for (k = 0, pSignal = Block::m_Signaux; k < Block::m_uiSignaux; ++k, ++pSignal) {
			if (pSignal->m_fModifie[ch]) {
				ptx += sprintf(ptx, "%02X%02X%02X", k, pSignal->m_Feux, pSignal->m_Ralenti);
				pSignal->m_fModifie[ch] = false;
			}
		}

		// zones occupancy
		Zone* pZone;

		*ptx++ = ' ';
		*ptx++ = 'Z';

		for (k = 0, pZone = Block::m_Zones; k < Block::m_uiZones; ++k, ++pZone) {
			if (pZone->m_cotePair.m_train != LIBRE || pZone->m_fModifiee[ch]) {
				Train* pTrain = &Block::m_Trains[pZone->m_cotePair.m_train];
				byte side = pZone->m_coteImpair.m_train ? COTE_PAIR : UNIQUE;

				if (pZone->m_cotePair.m_train && pZone->m_cotePair.m_train <= DERNIER_ENGIN) {
					ptx += sprintf(ptx, "%02X%c%02X%02X%02X%c", k, side, pZone->m_cotePair.m_train, pTrain->m_wagons, pTrain->m_pasActuel, pTrain->m_sensActuel);
				} else if (pZone->m_cotePair.m_train < INCONNU) {
					ptx += sprintf(ptx, "%02X%c%02X%02X00u", k, side, pZone->m_cotePair.m_train, pTrain->m_wagons);
				} else if (!au && pZone->EstOccupee()) {
					ptx += sprintf(ptx, "%02X%cFF0000u", k, side);
				} else
					ptx += sprintf(ptx, "%02XU000000u", k);

				if (pZone->m_coteImpair.m_train) {
					pTrain = &Block::m_Trains[pZone->m_coteImpair.m_train];

					if (pZone->m_coteImpair.m_train && pZone->m_coteImpair.m_train <= DERNIER_ENGIN) {
						ptx += sprintf(ptx, "%02XO%02X%02X%02X%c", k, pZone->m_coteImpair.m_train, pTrain->m_wagons, pTrain->m_pasActuel, pTrain->m_sensActuel);
					} else if (pZone->m_coteImpair.m_train < INCONNU) {
						ptx += sprintf(ptx, "%02XO%02X%02X00u", k, pZone->m_coteImpair.m_train, pTrain->m_wagons);
					} else if (!au && pZone->EstOccupee()) {
						ptx += sprintf(ptx, "%02XOFF0000u", k);
					}
				}

				pZone->m_fModifiee[ch] = false;
			}
		}
	}
#endif
#endif
	*ptx = '\0';
}

// ex de trames:
//    au1
//    a3 s+ v8 f0-
//    cva0001 cvd0008 cvp1234

/* Protocole
 *
 * au1 = arret d'urgence, le booster est desactive
 * au0 = fin d'arret d'urgence
 *
 * a<1-99>   = adresse de la locomotive et selection du canal A de la souris
 * b<1-99>   = adresse de la locomotive et selection du canal B de la souris
 * C<1-99>   = adresse de la locomotive et selection du canal C de la souris
 * D<1-99>   = adresse de la locomotive et selection du canal D de la souris
 * s+        = sens 1 (pour le canal selectionne)
 * s-        = sens 0
 * v<0-100>  = vitesse en %
 * V+		 = marche à vue
 * V-		 = fin de marche à vue
 * f<0-28>+  = allumage fonction auxiliaire
 * f<0-28>-  = extinction fonction auxiliaire
 *
 * cva<1-1024> cvd<0-255> cvp<XXXX> = programation d'un decodeur de locomotive cva=adresse du CV, cvd=data cvp=code de protection a fixer dans ce programme afin d'interdir les programmation non autorisees
 *
 * x<1-511>.<0-7>+  = activation d'une sortie d'un decodeur d'accessoire
 * x<1-511>.<0-7>-  = desactivation d'une sortie d'un decodeur d'accessoire (normalement jamais utilise car l'activation de 0 desactive 1 et vice versa, ou alors si le decodeur est configure en impulsion, les sorties reviennent a 0 toutes seules)
 * y<1-2044>=<0-31> = afectation de la valeur 0-31 a un decodeur d'accessoire etendu
 *
 * o<0-5>+   = mise a 1 d'une sortie directe
 * o<0-5>-   = mise a 0 d'une sortie directe
 *
 * l<0-127>+ = allumage d'une led
 * l<0-127>- = extinction d'une led
 * c<0-127>+ = mode clignotant
 * c<0-127>- = mode fixe
 * h<0-127>- = phase normale
 * h<0-127>+ = phase inverse
 *
 * p<0-47>=<0-100>  = pwm (sur i2c)
 * s<0-47>=<50-250> = servo(sur i2c) = pulsation en 10us, neutre en 150*10us=1.5ms
 * s<0-47>-         = mettre le servo en position 0
 * s<0-47>+         = mettre le servo en position 1
 * w<0-47>=<0-255>  = vitesse de rotation des servos (0=max, nb de us / 250ms, ex: 1 = 1us/250ms = 4us/sec, 1000->1500us en 125s)
 * Le choix entre pwm et servo se regle au debut du fichier d ela centrale
 * Pour utiliser les sorties en 0/1, utiliser pwm0/100
 * La vitesse de rotation des servos, ainsi que les positions extremes sont reglables
 *
 * t<0-47>- = aiguillage en position direct (t=turnable, aiguillage en Americain)
 * t<0-47>/ = aiguillage en position devie
 * t<0-47>^ = inversion de la position de l'aiguillage
 * l'aiguillage peut etre de n'importe quel type(bobine, moteur, serv) et situe n'importe ou(en sortie de l'esp, sur un module i2c pca9685, sur un decodeur d'accessoire),
 * c'est a l'utilisateur de le definir dans ce programme. Il est egalement possible d'indiquer la duree de l'impulsion et inverser la commande
 * les rapport d et e sont mis a jour un efois la commande effectuee
 *
 * u<0-95>- = mettre a 0 la variable utilisateur
 * u<0-95>+ = mettre a 1 la variable utilisateur
 * u<0-95>^ = inverser la variable utilisateur
 *
 * ? = demande de l'emission de l'etat
 * au[0/1] ana/dcc iXXXXXXXXXXXXXXXXXXXXXXXX dXXXXXXXXXXXX eXXXXXXXXXXXX uXXXXXXXXXXXXXXXXXXXXXXXX cYY gXXXXXXX..
 * 01 2   3456    789012345678901234567890123456789012345678901234567890123456789012345678901234567890123456789012
 * 0                 1         2         3         4         5         6         7         8         9          10
 * X = codage en hexa 0-F pour transmettre 4 bits par caracter afin de ne pas surcharger le lien
 *     on aurait pu passer en binaire pour encore plus gagner, mais je voulais rester en ASCII
 * i = entree (0-95)
 * u = etat des commandes utilisateur
 * d = aiguillage en position directe (0-47)
 * e = aiguillage en position devie   (0-47)
 * c = convertisseur analogique numerique (souvent utilise pour mesurer le courant
 * YY = 0(0V), 99(1V)
 *
 * ! = information sur la centrale: ie: D17 v20170106a
 *
 * J<01-FE><01-FE><01-FE><01-FE> = joindre deux trains. Zone, nouveau train, train_pair, m_coteImpair.m_train
 * SS <01-FE><01-FF><01-FE>E|O|u<01-FE>E|O|u = scinder le train en deux. Le premier est côté pair, le second côté impair.
 * T<01-FE><00-64><00-64><00-64><00-64><01-1C>
 * V+ = passer le train en marche à vue
 * V- = fin du mode marche à vue
 * Z<00-FE><01-FE><00-FF><E|O|u> = positionnement initial d'un train dans une zone 
 */

// jmd>>

static unsigned int nbr(char** pp)
{
    unsigned int nb = 0;

    while (isDigit(**pp)) {
        nb = nb * 10 + **pp - '0';
        ++(*pp);
    }

    return nb;
}

unsigned int hex2(char** pp)
{
	unsigned int nb = 0;
	byte c;

	for (byte i = 2; i; i-- && **pp, ++(*pp)) {
		nb <<= 4;

		if (!isHexadecimalDigit(c = **pp))
			c = '0';

		nb += (isDigit(c)) ? c - '0' : toupper(c) - 'A' + 10;
	}

	return nb;
}

unsigned int hex4(char** pp)
{
	unsigned int nb = 0;
	byte c;

	for (byte i = 4; i; i-- && **pp, ++(*pp)) {
		nb <<= 4;

		if (!isHexadecimalDigit(c = **pp))
			c = '0';

		nb += (isDigit(c)) ? c - '0' : toupper(c) - 'A' + 10;
	}

	return nb;
}

// << jmd

static bool first_au_exit = true;

void parse_rx(byte cclient, char* rx)
{
	byte slot = 99;
	byte nb, is_num;
    unsigned int nbi;
    byte ch_abcd = 0; //canal a par defaut
    C_slot_ch* pslot;

    while (*rx) {
        // pre-traitement
        is_num = isDigit(rx[1]);

        // skip

        if (!*rx || *rx == '\n')
            break;
        else if (*rx == ' ')
            ++rx;
        else if (*rx == 'c' && rx[1] == 'a') // cabine_net v1
        {
            TRACER.println("cabine_net v1 no more supported !");

            while (1)
                delay(0);
        } else if (*rx == 's' && rx[1] == 'o') {
#if PROTOCOLE == D17PPV2 || PROTOCOLE == D17PPV3
			if (rx[2] == 'v' && isdigit(rx[3])) {
#if PROTOCOLE == D17PPV3
				uiClientProtocol[cclient] = rx[3] - '0';
				fUseBlock |= uiClientProtocol[cclient] > 2;
#endif
				TRACE1("Client %d use protocol v%c\n", cclient, rx[3]);
			} else {
#if PROTOCOLE == D17PPV3
				uiClientProtocol[cclient] = 1;
#endif
				TRACE1("Client %d use protocol v1\n", cclient);
			}
#endif
            break; // souris_d17_net v1
        }
        // locomotives
        else if (*rx == 'a' && rx[1] == 'u') {
            // Arrêt d'urgence
            au = (rx[2] == '1');
            rx += 3;
#if PROTOCOLE == D17PPV3
			// Stoppe tous les trains tant qu'ils ne sont pas configurés
            if (fUseBlock && !au && first_au_exit) {
                first_au_exit = false;

                for (Train* pTrain = &Block::m_Trains[0]; pTrain <= &Block::m_Trains[DERNIER_TRAIN]; ++pTrain)
                    pTrain->RegulerVitesse(0);
            }
#endif
        } else if (is_num && (*rx == 'a' || *rx == 'b'
#if PROTOCOLE == D17PPV2 || PROTOCOLE == D17PPV3
			|| (*rx >= 'A' && *rx <= 'D')
#endif
			)) {
            // canal a, b, c, d
			ch_abcd = tolower(*rx++) - 'a';
			unsigned int adr = nbr(&rx);
			souris_adr[cclient][ch_abcd] = adr;

			slot = 99;
			if (adr != ADR_ANA)
				if ((adr >= 1) && (adr <= 10239)) {
					slot = get_slot_for_adr(adr);

					if (slot == 99) {
						slot = get_free_or_oldest_slot();
						pslot = &slot_ch[slot];
						pslot->adr = adr;
						pslot->sens = 1; // 1=AV                       //UU: voir si on initialise ou pas
						pslot->vit = 0;                                //UU: voir si on demande a mettre a jour la loco avec les modified=1

						for (byte j = 0; j < 29; j++) 
							pslot->fct[j] = 0;
#if PROTOCOLE == D17PPV3
						Train* pTrain = Block::FindTrainByAddress(adr);

						if (pTrain)
							pslot->speedsteps = pTrain->m_pasDCC;
#endif
					} else
						pslot = &slot_ch[slot];

					set_first_slot(slot);
				}
		} else if (*rx == 's' && (rx[1] == '+' || rx[1] == '-')) {
            // sens
			if (slot != 99) {
				pslot->sens = (*++rx == '+');
#if PROTOCOLE == D17PPV3
				if (fUseBlock)
					Block::m_Trains[pslot->adr].ChangerSens(*rx);
#endif
			}

            if (pslot->adr == ADR_ANA)
                pwmsens = (*rx == '+');
            ++rx;
        } else if (is_num && *rx == 'v') {
            // vitesse
            ++rx;
            nb = (byte)nbr(&rx); //0-100%

			if (slot != 99) {
#if PROTOCOLE == D17PPV3
				if (uiClientProtocol[cclient] > 2) {
					pslot->vit = nb;	// 0-m_pasMax

					if (fUseBlock)
						Block::m_Trains[pslot->adr].ChangerVitesse(nb);
				} else
#endif
					pslot->vit = map(nb, 0, 100, 0, USER_SPEED_STEPS);	// 0-28 ou 0-126
			}

			if (pslot->adr == ADR_ANA) {
				pwm1024 = nb * 1024 * USER_PWM_FACTOR / 100 / 100;
				if (pwm1024 < 40) pwm1024 = 0;                // limitations aux extremites du a un bug de la fonction analogwrite
				if (pwm1024 > (1023 - 40)) pwm1024 = 1024;    // 1024 met la sortie a 1 en permanence
				//TRACER.print("pwm="); TRACER.print(pwm1024); Serial.print("<\n");                
			}
       } else if (is_num && *rx == 'f') {
            // fonctions
            ++rx;
            nb = (byte)nbr(&rx);

            while (nb <= 28) {
                if (*rx == '+') {
					if (slot != 99)
						dcc_set_fct(slot, nb, 1);

                    ++rx;
                    ++nb;
                } else if (*rx == '-') {
					if (slot != 99)
						dcc_set_fct(slot, nb, 0);

                    ++rx;
                    ++nb;
                } else
                    break;
            }
        } else if (*rx == 'c' && rx[1] == 'v') {
            // CV
            byte a = rx[2];
            rx += 4;
            nbi = nbr(&rx);
            if (a == 'a')
                cv_adr = nbi;
            else if (a == 'd')
                cv_dat = (byte)nbi;
            else if (a == 'p' && nbi == CV_PROG_CODE && boost == BOOST_DCC) {
                led_sign_of_life(1); //LED ON
                TRACER.printf(tx, "prog cv%d=%d\n", cv_adr, cv_dat);
                // programing sequence
                prog_cv(cv_adr, cv_dat);
				yield();
                // disable booster during 1s to reboot decoder
                digitalWrite(P_PIN, LOW);
                digitalWrite(S_PIN, LOW);
                led_sign_of_life(0);
                delay(1000); //UU: possible to check with can
                digitalWrite(P_PIN, HIGH);
                TRACER.println("prog done");
            }
		} else if (is_num && *rx == 'o') {	// entrees/sorties
            // direct out
            ++rx;
            nb = (byte)nbr(&rx);
            user_out(nb, *rx == '+');
            ++rx;
        } else if (is_num && *rx == 'l') {
            // led
            ++rx;
            nb = (byte)nbr(&rx);
#if USER_USE_MAX
			if (nb < IO_LED_NB)
                user_led(nb, *rx == '+');
#endif
            ++rx;
        } else if (is_num && *rx == 'c') {
            // mode clignotant
            ++rx;
            nb = (byte)nbr(&rx);
#if USER_USE_MAX
            if (nb < IO_LED_NB)
                user_led_cli(nb, *rx == '+');
#endif
            ++rx;
        } else if (is_num && *rx == 'h') {
            // led normale ou inverse
            ++rx;
            nb = (byte)nbr(&rx);
#if USER_USE_MAX
			if (nb < IO_LED_NB)
                user_led_pha(nb, *rx == '+');
#endif
            ++rx;
        } else if (is_num && *rx == 'p') {
            // pwm p<0-47>=<0-100>  = pwm (sur i2c)
            ++rx;
            nb = (byte)nbr(&rx);
            ++rx;
            nbi = nbr(&rx);
#if USER_USE_PCA
            if (nb < IO_PCA9685_OUT_NB)
                user_pwm_0_100(nb, nbi);
#endif
        } else if (is_num && *rx == 's') {
            // servos s<0-47>=<50-250> = servo(sur i2c) = pulsation en 10us, neutre en 150*10us=1.5ms
            //        s<0-47>-         = mettre le servo en position 0
            //        s<0-47>+         = mettre le servo en position 1
            //        s<0-47>/         = mettre le servo en position 1
            //        w<0-47>=<0-255>  = vitesse de rotation des servos (0=max, nb de us / 250ms, ex: 1 = 1us/250ms = 4us/sec, 1000->1500us en 125s)
            ++rx;
            nb = (byte)nbr(&rx);

            if (*rx == '=') {
                ++rx;
                nbi = nbr(&rx);
#if USER_USE_PCA
				if (nb < IO_PCA9685_OUT_NB)
                    user_servo_500_2500(nb, 10 * nbi);
#endif
            }
            /* desactive
				  else
				  {
					  if(nb < IO_PCA9685_OUT_NB)
					  {
						  if(*rx == '-') set_srv_50_250(nb, srv_pos_0[nb]);
						  if(*rx == '+') set_srv_50_250(nb, srv_pos_1[nb]);
						  if(*rx == '/') set_srv_50_250(nb, srv_pos_1[nb]);
					  }
					  ++rx;
				  }
			*/
        } else if (is_num && *rx == 'w') {
            // vitesse de rotation servo
            ++rx;
            nb = (byte)nbr(&rx);

            if (*rx == '=') {
                ++rx;
                nbi = nbr(&rx);
#if USER_USE_PCA
                if (nb < IO_PCA9685_OUT_NB)
                    user_servo_speed(nb, nbi);
#endif
            }
        } else if (is_num && *rx == 'x') {
            // decodeur d'accesssoire "basique"
            // x<1-511>.<0-7>+ = activation d'une sortie d'un decodeur d'accessoire
            // x<1-511>.<0-7>- = desactivation d'une sortie d'un decodeur d'accessoire
            //                   (normalement jamais utilise car l'activation de la sortie 0 desactive la 1 et vice versa,
            //                   ou alors si le decodeur est configure en impulsion, les sorties reviennent a 0 toutes seules)
            ++rx;
            nbi = nbr(&rx);
            ++rx;
            nb = (byte)nbr(&rx);
            //TRACE1("x=%d", nbi);
            //TRACE1(" out=%d", nb);
            if (nbi >= 1 && nbi <= 511 && nb <= 7)
                user_bas_acc_dec_tx(nbi, nb, *rx == '-');
            ++rx;
        } else if (is_num && *rx == 'y') {
            // decodeur d'accesssoire "extended"
            // y<1-2044>=<0-31> activation d'une sortie d'un decodeur d'accessoire
            ++rx;
            nbi = nbr(&rx);
            ++rx;
            nb = (byte)nbr(&rx);
            if (nbi >= 1 && nbi <= 2044 && nb <= 31)
                user_ext_acc_dec_tx(nbi, nb);
        } else if (is_num && *rx == 't') {
            // aiguillages
            // t<0-47>- = aiguillage en position direct (t=turnable, aiguillage en Americain)
            // t<0-47>/ = aiguillage en position devie
            // t<0-47>^ = inversion de la position de l'aiguillage
            ++rx;
            nb = (byte)nbr(&rx);

            if (nb < IO_AIG_NB) {
                if (*rx == '-')
                    io_aig_cmd[nb] = 0;
                if (*rx == '+')
                    io_aig_cmd[nb] = 1;
                if (*rx == '/')
                    io_aig_cmd[nb] = 1;
                if (*rx == '^')
                    io_aig_cmd[nb] ^= 1;
#if PROTOCOLE == D17PPV3
				if (fUseBlock)
					Block::OnNotifyAig(nb, io_aig_cmd[nb]);
				else
					on_notify_aig(nb, io_aig_cmd[nb]);
#else
                on_notify_aig(nb, io_aig_cmd[nb]);
#endif
            }
            ++rx;
        } else if (is_num && *rx == 'u') {
            // user variables
            // u<0-95>- = mettre a 0 la variable utilisateur
            // u<0-95>+ = mettre a 1 la variable utilisateur
            // u<0-95>^ = inverser la variable utilisateur
            ++rx;
            nb = (byte)nbr(&rx);

            if (nb < VAR_U_NB) {
                if (*rx == '-')
                    set_var_u(nb, 0);
                if (*rx == '+')
                    set_var_u(nb, 1);
                if (*rx == '/')
                    set_var_u(nb, 1);
                if (*rx == '^')
                    inv_var_u(nb);
#if PROTOCOLE == D17PPV3
				if (fUseBlock)
					Block::OnNotifyU(nb, get_var_u(nb));
				else
					on_notify_u(nb, get_var_u(nb));
#else
                on_notify_u(nb, get_var_u(nb));
#endif
            }
            ++rx;
        } else if (*rx == '?') {	// query
            // status
            ++rx;

            if (status_tx[0] == '\0') {
                build_status_message(cclient);
                status_txed[0] = status_txed[1] = status_txed[2] = status_txed[3] = 0;
            }
            if (status_txed[cclient] == 0) {
                update_in_status_message(cclient);
                strcpy(tx, status_tx);
#if USE_TRACER == 1 && PRINT_STATUS
                if (strcmp(status_tx_old, status_tx)) {
                    TRACER.println(tx);
                    strcpy(status_tx_old, status_tx);
                }
#endif
                status_txed[cclient] = 1;
            }
        } else if (*rx == '!') {
            // version
            ++rx;
            strcpy(tx, CENTRAL_NAME);
            TRACER.printf("%s\n", tx);
#if PROTOCOLE == D17PPV2 || PROTOCOLE == D17PPV3
		} else if (*rx == 'P') {	// CV programming
			unsigned int address;
			byte bit;
			byte value = 0;
			byte mode;

			++rx;
			Decoder::m_fResult = true;

			switch ((Decoder::m_cCode = *rx++)) {
			case 'i':	// Reset decoder
				Decoder::Reset(3);
				break;
			case 'F':	// reset decoder to factory default
				Decoder::Factory();
				break;
			case 'R':	// PR : DirectRead CV byte : address
				mode = *rx++;
				address = hex4(&rx);

				if (mode == 'D')
					Decoder::DirectRead(address);
				else
					Decoder::PagedRead(address);
				break;
			case 'r':	// Pr : DirectReadBit CV bit : address, bit
				address = hex4(&rx);
				Decoder::DirectReadBit(address, hex2(&rx));
				break;
			case 'V':	// PV : DirectVerify CV byte : address, value
				mode = *rx++;
				address = hex4(&rx);

				if (mode == 'D')
					Decoder::DirectVerify(address, hex2(&rx));
				else
					Decoder::PagedVerify(address, hex2(&rx));
				break;
			case 'W':	// PW : DirectWrite CV byte : address, value
				mode = *rx++;
				address = hex4(&rx);

				if (mode == 'D')
					Decoder::DirectWrite(address, hex2(&rx));
				else
					Decoder::PagedWrite(address, hex2(&rx));

				// disable booster during 1s to reboot decoder
				digitalWrite(P_PIN, LOW);
				digitalWrite(S_PIN, LOW);
				led_sign_of_life(0);
				delay(1000);
				digitalWrite(P_PIN, HIGH);
				break;
			case 'w':	// Pw : DirectWriteBit CV bit : address, bit number, value
				address = hex4(&rx);
				bit = hex2(&rx);
				Decoder::DirectWriteBit(address, bit, hex2(&rx) != 0);
				break;
			}
		} else if (is_num && *rx == 'M') {
			// Mesurer le temps de boucle d'un train : M <01-7F> index, <01-237F> adresse, <00-1C|7E> pas dcc, <00-FE> zone de mesure, <0-1> tour de lancement.
			// Le résultat en ms est envoyé dans la prochaine chaîne de statut.
			// Pendant la mesure tous les automatismes de sécurité sont désactivés. Un seul train doit donc être en circulation.
			++rx;
			byte train = hex2(&rx);
			unsigned int adresse = hex4(&rx);
			byte steps = hex2(&rx);
			byte pas = hex2(&rx);
			byte zone = hex2(&rx);
			byte tour = hex2(&rx);
			
			if (zone < 
#if PROTOCOLE == D17PPV3
				Block::m_uiZones && train <= DERNIER_TRAIN
#else
				IO_IN_NB
#endif
				)
				Speed::MesureVitesse(train, adresse, steps, pas, zone, tour);
		} else if (is_num && *rx == 'R') {
			if (*++rx == '1') {
				TRACE1("Restart D17\n");
				TRACER.flush();
				ESP.restart();
			}

			++rx;
#if PROTOCOLE == D17PPV3
		} else if (fUseBlock) {
			if (is_num && *rx == 'Z') {
				// initialisation zone : Z, <00-FE> zone, U|O|E côté, <01-FE> train, <00-FF> wagons, e|o|u sens
				++rx;
				byte zone = hex2(&rx);
				byte train_pair = hex2(&rx);
				byte wagons_pair = hex2(&rx);
				byte sens_pair = *rx++;
				byte train_impair = hex2(&rx);
				byte wagons_impair = hex2(&rx);
				byte sens_impair = *rx++;

				if (zone < Block::m_uiZones)
					Block::m_Zones[zone].InitTrains(train_pair, wagons_pair, sens_pair, train_impair, wagons_impair, sens_impair);
			} else if (is_num && *rx == 'J') {
				// joindre deux trains J <01-FE> zone, <01-FE> train, e|o|u sens
				++rx;
				byte zone = hex2(&rx);
				byte train = hex2(&rx);
				byte sens = *rx++;

				if (zone < Block::m_uiZones)
					Block::m_Zones[zone].JoinTrains(train, sens);
			} else if (is_num && *rx == 'S') {
				// séparer en deux trains S <00-FE> zone, <01-FF> coupure_avant, <01-FE> nouveau_train_pair, e|o|u sens, <01-FE> nouveau_train_impair, e|o|u sens
				++rx;
				byte zone = hex2(&rx);
				byte coupure_avant = hex2(&rx);
				byte train_pair = hex2(&rx);
				byte sens_pair = *rx++;
				byte train_impair = hex2(&rx);
				byte sens_impair = *rx++;

				if (zone < Block::m_uiZones)
					Block::m_Zones[zone].SplitTrain(coupure_avant, train_pair, sens_pair, train_impair, sens_impair);
			} else if (is_num && *rx == 'T') {
				// configuration d'un engin moteur : <01-7F> index, <0001-27FF> adresse dcc, <0000-FFFF> vitesse maxi réelle, <0000-FFFF> vitesse max mesurée, <01-1C> fontion manoeuvre,
				//									 <00-64> 15 km/h, <00-64> 15 km/h, <00-64> 30 km/h, <00-64> 90 km/h, <00-64> 160 km/h
				++rx;
				byte index = hex2(&rx);
				unsigned int adresse = hex4(&rx);
				byte speedsteps = hex2(&rx);
				byte vmanoeuvre = hex2(&rx);
				byte vitesse15 = hex2(&rx);
				byte vitesse30 = hex2(&rx);
				byte vitesse60 = hex2(&rx);
				byte vitesse90 = hex2(&rx);
				byte vitesse100 = hex2(&rx);
				byte vitesse120 = hex2(&rx);
				byte vitesse160 = hex2(&rx);
				byte vitesse220 = hex2(&rx);
				byte vitesse300 = hex2(&rx);
				byte vitesse320 = hex2(&rx);
				byte vitesseMax = hex2(&rx);

				if (index <= DERNIER_ENGIN)
					Block::m_Trains[index] = Train(adresse, speedsteps, vmanoeuvre,
						vitesse15, vitesse30, vitesse60, vitesse90, vitesse100, vitesse120, vitesse160, vitesse220, vitesse300, vitesse320, vitesseMax);	// train avec engin moteur
			} else if (is_num && *rx == 'V') {
				// marche à vue
				++rx;
				Block::m_Trains[pslot->adr].ChangerMarcheAVue(nbr(&rx));
			}
#endif
#endif
		} else
            ++rx;

        yield();
    }
}

#endif