#include "speed.h"

#if PROTOCOLE == D17PPV2 || PROTOCOLE == D17PPV3

bool Speed::m_fEnMesure = false;	// true si mesure en cours.
bool Speed::m_fMesureFaite = false;	// true si mesure termine.
C_slot_ch* Speed::m_pSlot;			// Slot du train.
byte Speed::m_ucTrain;				// Index du train dans la table des trains
byte Speed::m_ucPas;				// Pas de vitesse DCC
byte Speed::m_ucZone;				// Zone de mesure.
byte Speed::m_uiPhase;				// Phase de la mesure.
unsigned long Speed::m_uiStart;		// "Heure" de dbut de mesure.
unsigned long Speed::m_uiTemps;		// Temps de parcours en ms.
unsigned long Speed::m_uiFreed;		// "Heure" de libration de la zone pour l'anti rebond.

void Speed::MesureVitesse(byte train, unsigned int adresse, byte speedsteps, byte pas, byte zone, byte tour)
{
	byte slot = get_slot_for_adr(adresse);

	if (slot == 99)
		slot = get_free_or_oldest_slot();

	TRACE1("Speed::MesureVitesse train %d, slot %d, step %d, zone %d, loop %d\n", adresse, slot, pas, zone, tour);
	m_pSlot = &slot_ch[slot];
	m_ucTrain = train;
	m_ucPas = pas;
	m_ucZone = zone;
	m_fEnMesure = true;
	m_uiPhase = tour ? 0 : 2;
	m_pSlot->adr = adresse;
	m_pSlot->vit = pas;
	m_pSlot->sens = 1;
#if PROTOCOLE == D17PPV3
	m_pSlot->speedsteps = speedsteps;
#endif
}

void Speed::MesureVitesse(void)
{
	switch (m_uiPhase) {
	case 0: // Attente premier passage tour de lancement.
		if (user_get_in(m_ucZone)) {
			TRACE1("Speed::MesureVitesse premier passage tour de lancement\n");
			m_uiPhase = 1;
		} else
			yield();

		break;
	case 1:	// Attente libration zone
		if (!user_get_in(m_ucZone)) {
			TRACE1("Speed::MesureVitesse zone liberee tour de lancement\n");
			m_uiPhase = 2;
			m_uiFreed = millis();
		} else
			yield();

		break;
	case 2:	// Attente premier passage
		if (user_get_in(m_ucZone) && millis() - m_uiFreed > 2000) {
			TRACE1("Speed::MesureVitesse premier passage\n");
			m_uiPhase = 3;
			m_uiStart = millis();
		} else
			yield();

		break;
	case 3:	// Attente libration zone
		if (!user_get_in(m_ucZone)) {
			TRACE1("Speed::MesureVitesse zone liberee\n");
			m_uiPhase = 4;
			m_uiFreed = millis();
		}  else
			yield();

		break;
	case 4:	// Attente second passage.
		if (user_get_in(m_ucZone) && millis() - m_uiFreed > 2000) {
			TRACE1("Speed::MesureVitesse second passage\n");
			m_uiTemps = millis() - m_uiStart;
			m_fEnMesure = false;
			m_fMesureFaite = true;
			m_pSlot->vit = 0;
		} else
			yield();

		break;
	}
}

#endif