//--------------------------------------------------------------------------------
// Module D17 par Ulysse Delmas-Bgue et  la marge Jean-Michel Dubois
// TCP/IP server part
//----------------------------------------------------------------------------------
#include "config.h"

#if USE_WIFI || USE_TRACER == 2

#include "tcpip.h"
#include "dcc.h"
#include "protocol.h"
#include "block.h"
#include "signal.h"

#if USE_WIFI
static WiFiServer server(TCP_PORT);
WiFiClient zeclients[MAX_WIFI_CLIENTS];
static byte wait_id_msg[MAX_WIFI_CLIENTS];
#endif
#if USE_TRACER == 2
WiFiServer tracer(TRACER_PORT);
WiFiClient trclient;
#endif

void wifi_init(void)
{
// jmd>>
#if APMODE
    WiFi.mode(WIFI_AP);
    WiFi.softAP(WIFI_SSID, WIFI_PASS);
# if DEFAULT_AP_CONFIG
    IPAddress myIP = WiFi.softAPIP();
# else
    IPAddress myIP IPADDRESS;
    IPAddress gateway GATEWAY;
    IPAddress subnet SUBNET;
    WiFi.softAPConfig(myIP, gateway, subnet);
# endif
#else
    IPAddress myIP IPADDRESS;
    IPAddress gateway GATEWAY;
    IPAddress subnet SUBNET;
    WiFi.config(myIP, gateway, subnet);
    WiFi.mode(WIFI_STA);
    WiFi.begin(WIFI_SSID, WIFI_PASS);
#endif
	Serial.print("D17 IP address: ");
	Serial.println(myIP);
#if USE_WIFI
    server.begin();
	memset(wait_id_msg, 0, sizeof(wait_id_msg));
#if PROTOCOLE <= D17PPV3
	memset(status_txed, 0, sizeof(status_txed));
#endif
	memset(souris_adr, 0, sizeof(souris_adr));
#endif
#if USE_TRACER == 2
	tracer.begin();
#endif
}

#if USE_TRACER
static String lineo; // jmd
#endif

#if USE_WIFI

void parse_wifi_clients(void)
{
    byte i;
    char rx[1024] = "";

  
    for(i = 0; i < MAX_WIFI_CLIENTS; i++)
    {
        // Client comes
        if(!zeclients[i]) 
        { 
            zeclients[i] = server.available();
            if(zeclients[i]) 
            {
                Serial.printf("Client %d connected\n", i);
#if PROTOCOLE == D17PPV3
                zeclients[i].print("d17 v3\n");
#elif PROTOCOLE == D17PPV2
                zeclients[i].print("d17 v2\n");
#else
                zeclients[i].print("d17 v1\n");
#endif
				wait_id_msg[i] = 1;
#if PROTOCOLE == D17PPV3
				Signal* pSignal;
				byte k;

				for (k = 0, pSignal = Block::m_Signaux; k < Block::m_uiSignaux; ++k, ++pSignal)
					pSignal->m_fModifie[i] = true;
#endif
			}
        }
        
        // No Client
        if(!zeclients[i]) continue;
        
        // Client leaves
        if(!zeclients[i].connected()) 
        {
			Serial.printf("Client %d disconnected\n", i);
			zeclients[i].stop();
#if PROTOCOLE == D17PPV3
			uiClientProtocol[i] = 0;
#endif
            //zeclients[i] = NULL;  //uu: essayer
            continue;
        }

        // Client sends data
        while(zeclients[i].available()) 
        {
            String line = zeclients[i].readStringUntil('\n');
            line.toCharArray(rx, 119);
            rx[119] = '\0';

#if USE_TRACER && PRINT_COMMAND
            if (lineo != line && *rx) {  // jmd
                TRACE1("Client %d '%s'\n", i, rx);
                lineo = line;     // jmd
            }                     // jmd
#endif
            yield();

			if (*rx) // parse the command message
            {
                parse_rx(i, rx); 

				if(tx[0] != '\0') 
                {
                     // envoie du message de reponse
                     strcat(tx, "\n"); 
                     zeclients[i].print(tx);
                     tx[0] = '\0'; 
                }
            }

            yield();
// jmd            TRACE1(".");
        }
        //client.flush();
    }

    delay(0);      
}

#endif

#if USE_TRACER == 2

void connect_tracer(void)
{
	if (!trclient)
	{
		if (trclient = tracer.available())
		{
#if PROTOCOLE == D17PPV3
		trclient.print("d17++ v3\r\n");
#elif PROTOCOLE == D17PPV2
		trclient.print("d17++ v2\r\n");
#elif PROTOCOLE == D17
		trclient.print("d17 v1\r\n");
#elif PROTOCOLE == DCCPP
		trclient.print("protocole DCC++ pour Rocrail\r\n");
#elif PROTOCOLE == CDMRAIL
		trclient.print("protocole DCC++ pour CDM-RAIL\r\n");
#elif PROTOCOLE == DSTATION
		trclient.print("protocole Desktop Station\r\n");
#endif
		}
	}
	else if (!trclient.connected())
		trclient.stop();
}

#endif

#endif